#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
import copy
import logging
import os
import re
import sys
import time
from tkinter import messagebox

from geofinder import DB, Loc, GeoKeys, Geodata
from geofinder.GeoKeys import Query, Result, Entry, get_soundex


class GeoDB:
    """
    geoname data database.  Add items, look up items, create tables, indices
    """

    def __init__(self, database):
        self.logger = logging.getLogger(__name__)
        self.start = 0

        self.database = database
        # See if DB exists
        if os.path.exists(database):
            db_exists = True
        else:
            db_exists = False

        self.db = DB.DB(database)
        if self.db.err:
            self.logger.error(f"Error! cannot open database {database}.")
            raise ValueError('Cannot open database')

        # If DB exists
        if db_exists:
            # Run sanity test on DB
            res = self.db.db_test('main.geodata')

            if res:
                self.logger.debug(f'DB error for {database}')
                if messagebox.askyesno('Error',
                                       f'Geoname database is empty or corrupt:\n\n {database} \n\nDo you want to delete it and rebuild?'):
                    messagebox.showinfo('', 'Deleting Geoname database')
                    self.db.conn.close()
                    os.remove(database)

                sys.exit()
        else:
            # DB didnt exist.  Create tables.
            self.create_tables()

        self.db.set_speed_pragmas()
        self.db.set_params(select_str='name, country, admin1_id, admin2_id, lat, lon, f_code, geoid, sdx', order_str='', limit_str='LIMIT 160')
        self.geoid_main_dict = {}  # Key is GEOID, Value is DB ID for entry
        self.geoid_admin_dict = {}  # Key is GEOID, Value is DB ID for entry
        self.place_type = ''

    def lookup_place(self, place: Loc.Loc) -> []:
        """
        Lookup a place in our geoname.org dictionary and update place with Geo_result with lat, long, District, etc
        The dictionary geo_result entry contains: Lat, Long, districtID (County or State or Province ID)
        There can be multiple entries if a city name isnt unique in a country
        """
        result_place: Loc = Loc.Loc()
        self.start = time.time()
        place.result_type = Result.STRONG_MATCH

        # Lookup Place based on Place Type
        if place.place_type == Loc.PlaceType.ADMIN1:
            self.select_admin1(place)
        elif place.place_type == Loc.PlaceType.ADMIN2:
            self.get_admin1_id(place=place)
            self.select_admin2(place)
            if len(place.georow_list) ==0:
                if 'shire' in place.target:
                    # Some English counties have had Shire removed from name, try doing that
                    place.target = re.sub('shire', '', place.target)
                    self.select_admin2(place)

        elif place.place_type == Loc.PlaceType.COUNTRY:
            self.select_country(place)
        elif place.place_type == Loc.PlaceType.ADVANCED_SEARCH:
            self.advanced_search(place)
        else:
            self.get_admin2_id(place=place)
            self.select_city(place)

        nm = place.name

        # Add search quality score to each entry
        for idx, rw in enumerate(place.georow_list):
            self.copy_georow_to_place(row=rw, place=result_place)
            update = list(rw)
            update.append(1)  # Extend list row and assign score
            result_place.prefix = ''
            res_nm = result_place.format_full_nm(None)
            # todo - move feature priority into scoring routine
            score = self.match_score(inp=nm, res=res_nm,
                                     feature_score=Geodata.Geodata.get_priority(rw[GeoKeys.Entry.FEAT]))

            # Remove items in prefix that are in result
            tk_list = res_nm.split(",")
            for item in tk_list:
                place.prefix = re.sub(item.strip(' ').lower(), '', place.prefix)

            update[GeoKeys.Entry.SCORE] = score
            place.georow_list[idx] = tuple(update)

        if place.result_type == Result.STRONG_MATCH and len(place.prefix) > 0:
            place.result_type = Result.PARTIAL_MATCH

    def match_score(self, inp, res, feature_score)->int:
        inp = GeoKeys.semi_normalize(inp)
        res = GeoKeys.semi_normalize(res)
        res = re.sub(r"'", ' ', res)  # Normalize
        res = GeoKeys.remove_noise_words(res)

        score : int = self.match_score_calc(inp, res)
        return  score + feature_score

    def match_score_calc(self, inp, res)->int:
        # Return a score 0-100 reflecting the difference between the user input and the result:
        # The percent of characters in inp that were NOT matched by a word in result
        # Lower score is better match.  0 is perfect match, 100 is no match

        original_inp_tokens = inp.split(',')
        inp2 = copy.copy(inp)
        inp2 = re.sub(',', ' ', inp2)
        inp_words = inp2.split(' ')

        result_tokens = res.split(",")

        # Counties are frequently wrong.  Minimize impact
        if len(result_tokens) > 2:
            # See if result county is in user request.  If it is not, delete it
            if result_tokens[-3].strip(' ') not in inp:
                result_tokens[-3] = ''

        # Countries are frequently left out.  Minimize impact
        if len(result_tokens) > 0:
            # See if result country is in user request.  If it is not, delete it
            if result_tokens[-1].strip(' ') not in inp:
                result_tokens[-1] = ''

        # Join tokens back into comma separated list
        out = ','.join(map(str, result_tokens))
        orig_res_len = len(out)

        # If any WORD in input is in result, delete it from input and from result
        lst = sorted(inp_words, key=len, reverse=True)      # Use longest words first
        for idx, word in enumerate(lst):
            if len(word) > 2 and word in res:
                # If input word is in result, delete it from input string and result string
                inp = re.sub(word, '', inp)
                out = re.sub(word, '', out)

        in_score = 0

        # Strip out any remaining CHARACTERS in inp that match out
        length = min(len(out), len(inp))
        match_list = [i for i in range(length) if inp[i] == out[i]]
        char_list = list(inp)    # Convert string to list so we can modify
        for idx in match_list:
            char_list[idx] = ' '
        inp = ''.join(char_list)   # Convert back from list to string

        #self.logger.debug(f' In [{inp}] Res [{out}]')

        # For each input token calculate new size divided by original size
        inp_tokens = inp.split(',')

        for idx, tk in enumerate(inp_tokens):
            # Tokens to the right end have slightly higher weighting
            weight = (1.0 + idx * .005)
            if len(original_inp_tokens[idx].strip(' ')) > 0:
                in_score += int(100.0 * len(inp_tokens[idx].strip(' ')) / len(original_inp_tokens[idx].strip(' ')) * weight)

        # Average over number of tokens, e.g. average percent of tokens unmatched
        in_score = in_score / len(original_inp_tokens)

        # Output score (percent of first token in output that was not matched)
        new_len = len(out.strip(' '))
        if orig_res_len > 0:
            out_score = (int((0.09 * 100.0) * new_len / orig_res_len))
            if new_len > 0:
                out_score += 1
        else:
            out_score = 0

        score = in_score + out_score
        #self.logger.debug(f'Sc={score}  [{original_inp_tokens[0]}] DB [{res}]  InS={in_score:.1f} InRem [{inp}] '
        #                  f'OutS={out_score:.1f} OutRem [{out}]  ')
        return int(score)

        """
        res:float = float(Levenshtein.distance(str(GeoKeys.normalize(inp)), str(GeoKeys.normalize(result))))
        if len(inp) > 0:
            return int(100*res / float(len(inp)))
        else:
            if len(result) > 0:
                return 100
            else:
                return 0
        """

    def advanced_search(self, place: Loc):
        """
        Advanced search - support parameters for ISO and Feature class
        """
        lookup_target = place.target
        if len(lookup_target) == 0:
            return
        pattern = self.create_wildcard(lookup_target)
        iso_pattern = place.country_iso
        feature_pattern = self.create_wildcard(place.feature)
        self.logger.debug(f'Advanced Search. Targ=[{pattern}] feature=[{feature_pattern}]'
                          f'  iso=[{iso_pattern}] ')

        query_list = []
        query_list.append(Query(where="name LIKE ? AND country LIKE ? AND f_code LIKE ?",
                                args=(pattern, iso_pattern, feature_pattern),
                                result=Result.PARTIAL_MATCH))

        # Search main DB
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.geodata', query_list=query_list)
        self.logger.debug(place.georow_list)

        # Search admin DB
        admin_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)
        place.georow_list.extend(admin_list)
        self.logger.debug(place.georow_list)

    def select_city(self, place: Loc):
        """
        Search for  entry - try the most exact match first, then less exact matches
        """
        lookup_target = place.target
        if len(lookup_target) == 0:
            return
        pattern = self.create_wildcard(lookup_target)
        sdx = get_soundex(lookup_target)
        self.logger.debug(f'CITY lookup. Targ=[{lookup_target}] adm1 id=[{place.admin1_id}]'
                          f' adm2 id=[{place.admin2_id}] iso=[{place.country_iso}] patt =[{pattern}] sdx={sdx}')

        query_list = []

        if len(place.country_iso) == 0:
            # No country present - try lookup by name.
            query_list.append(Query(where="name = ?",
                                    args=(lookup_target,),
                                    result=Result.PARTIAL_MATCH))
            # lookup by wildcard name, admin1
            query_list.append(Query(where="name LIKE ?",
                                    args=(pattern,),
                                    result=Result.WILDCARD_MATCH))
            #query_list.append(Query(where="name LIKE ?",
            #                        args=(pattern + ' historical',),
            #                        result=Result.WILDCARD_MATCH))
            query_list.append(Query(where="sdx = ?",
                                    args=(sdx,),
                                    result=Result.SOUNDEX_MATCH))
            place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.geodata', query_list=query_list)
            # self.logger.debug(place.georow_list)
            return

        # Build query list - try each query in order until a match is found
        # Start with the most exact match depending on the data provided.
        query_list.append(Query(where="name = ? AND country = ?",
                                args=(lookup_target, place.country_iso),
                                result=Result.PARTIAL_MATCH))

        if len(place.admin1_name) > 0:
            if len(place.admin2_name) > 0:
                # lookup by name, admin1, admin2
                query_list.append(Query(where="name = ? AND country = ? AND admin1_id = ? AND admin2_id = ?",
                                        args=(lookup_target, place.country_iso, place.admin1_id, place.admin2_id),
                                        result=Result.STRONG_MATCH))

            # lookup by name, admin1
            query_list.append(Query(where="name = ? AND country = ? AND admin1_id = ?",
                                    args=(lookup_target, place.country_iso, place.admin1_id),
                                    result=Result.STRONG_MATCH))

            # lookup by wildcard name, admin1
            query_list.append(Query(where="name LIKE ? AND country = ? AND admin1_id = ?",
                                    args=(pattern, place.country_iso, place.admin1_id),
                                    result=Result.PARTIAL_MATCH))
        elif len(place.admin2_name) == 0:
            # lookup by name
            # admin1 and admin2 werent entered, so a match here is an exact match
            query_list.append(Query(where="name = ? AND country = ?",
                                    args=(lookup_target, place.country_iso),
                                    result=Result.STRONG_MATCH))
        elif len(place.admin2_name) > 0:
            # lookup by name
            # admin1 wasnt entered, so a match here is an exact match
            query_list.append(Query(where="name = ? AND admin2_id=? AND country = ?",
                                    args=(lookup_target, place.admin2_id, place.country_iso),
                                    result=Result.STRONG_MATCH))
        else:
            # lookup by name (partial match since user specified admin)
            query_list.append(Query(where="name = ? AND country = ?",
                                    args=(lookup_target, place.country_iso),
                                    result=Result.PARTIAL_MATCH))

        # append lookup by wildcard (partial match since user specified admin)
        query_list.append(Query(where="name LIKE ? AND country = ?",
                                args=(pattern, place.country_iso),
                                result=Result.WILDCARD_MATCH))

        # lookup by wildcard with % added (partial match since user specified admin)
        query_list.append(Query(where="name LIKE ? AND country = ?",
                                args=('%' + pattern, place.country_iso),
                                result=Result.WILDCARD_MATCH))

        # lookup by Soundex (partial match)
        if len(place.admin1_name) > 0:
            query_list.append(Query(where="sdx = ? AND admin1_id = ? AND country = ?",
                                    args=(sdx, place.admin1_id, place.country_iso),
                                    result=Result.SOUNDEX_MATCH))
        else:
            query_list.append(Query(where="sdx = ? AND country = ?",
                                    args=(sdx, place.country_iso),
                                    result=Result.SOUNDEX_MATCH))

        # Try each query in list until we find a match
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.geodata', query_list=query_list)
        # self.logger.debug(place.georow_list)

    def select_admin2(self, place: Loc):
        """Search for Admin2 entry"""
        lookup_target = place.admin2_name
        if len(lookup_target) == 0:
            return
        place.target = lookup_target
        # sdx = get_soundex(lookup_target)

        # Try Admin query until we find a match - each query gets less exact
        query_list = [
            Query(where="name = ? AND country = ? AND admin1_id = ? AND f_code=?",
                  args=(lookup_target, place.country_iso, place.admin1_id, 'ADM2'),
                  result=Result.STRONG_MATCH),
            Query(where="name = ? AND country = ? AND f_code=?",
                  args=(lookup_target, place.country_iso, 'ADM2'),
                  result=Result.PARTIAL_MATCH),
            Query(where="name LIKE ? AND country = ? AND f_code=?",
                  args=(self.create_wildcard(lookup_target), place.country_iso, 'ADM2'),
                  result=Result.PARTIAL_MATCH),
            Query(where="name = ?  AND f_code=?",
                  args=(lookup_target, 'ADM2'),
                  result=Result.PARTIAL_MATCH),
            Query(where="name LIKE ? AND country = ? AND f_code=?",
                  args=(self.create_county_wildcard(lookup_target), place.country_iso, 'ADM2'),
                  result=Result.WILDCARD_MATCH)
        ]

        # self.logger.debug(f'Admin2 lookup=[{lookup_target}] country=[{place.country_iso}]')
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)
        if place.result_type == GeoKeys.Result.WILDCARD_MATCH:
            # Found as Admin2 without shire
            place.name = re.sub('shire', '', place.name)

        if len(place.georow_list) == 0:
            # Try city rather than County match.
            save_admin2 = place.admin2_name
            place.city1 = place.admin2_name
            place.admin2_name = ''
            # self.logger.debug(f'Try admin2 as city: [{place.target}]')

            self.select_city(place)

            if len(place.georow_list) == 0:
                #  not found.  restore admin
                place.admin2_name = save_admin2
                place.city1 = ''
            else:
                # Found match as a City
                place.place_type = Loc.PlaceType.CITY
                match_adm1 = self.get_admin1_name_direct(lookup_target=place.georow_list[0][Entry.ADM1], iso=place.country_iso)
                # self.logger.debug(f'pl_iso [{place.country_iso}] pl_adm1 {place.admin1_name} match_adm1=[{match_adm1}] ')
                if place.admin1_name != match_adm1:
                    place.prefix = place.admin1_name.title()
                    place.admin1_name = ''
                return

    def select_admin1(self, place: Loc):
        """Search for Admin1 entry"""
        lookup_target = place.admin1_name

        pattern = self.create_wildcard(lookup_target)
        if len(lookup_target) == 0:
            return
        sdx = get_soundex(lookup_target)

        # self.logger.debug(f'sel adm1 patt={pattern} iso={place.country_iso}')

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="name = ? AND country = ? AND f_code = ? ",
                  args=(lookup_target, place.country_iso, 'ADM1'),
                  result=Result.STRONG_MATCH),
            Query(where="name LIKE ? AND country = ?  AND f_code = ?",
                  args=(pattern, place.country_iso, 'ADM1'),
                  result=Result.WILDCARD_MATCH),
            Query(where="sdx = ? AND country = ? AND f_code=?",
                  args=(sdx, place.country_iso, 'ADM1'),
                  result=Result.SOUNDEX_MATCH)
        ]
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        """
        # if nothing found, Try admin1 as city instead
        if len(place.georow_list) == 0:
            save_admin1 = place.admin1_name
            place.city1 = place.admin1_name
            place.target = place.city1
            place.admin1_name = ''
            self.logger.debug(f'Try admin1: [{place.target}]')

            self.select_city(place)

            if len(place.georow_list) == 0:
                # still not found.  restore admin1
                place.admin1_name = save_admin1
                place.city1 = ''
            else:
                place.place_type = Loc.PlaceType.CITY
        """

    def select_country(self, place: Loc):
        """Search for Admin1 entry"""
        lookup_target = place.country_iso
        if len(lookup_target) == 0:
            return
        sdx = get_soundex(lookup_target)

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="country = ? AND f_code = ? ",
                  args=(place.country_iso, 'ADM0'),
                  result=Result.STRONG_MATCH),
            Query(where="sdx = ?  AND f_code=?",
                  args=(sdx, 'ADM0'),
                  result=Result.SOUNDEX_MATCH)
        ]

        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)
        # self.build_result_list(place.georow_list, place.event_year)

    def get_admin1_id(self, place: Loc):
        """Search for Admin1 entry"""
        lookup_target = place.admin1_name
        if len(lookup_target) == 0:
            return

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="name = ? AND country = ? AND f_code = ? ",
                  args=(lookup_target, place.country_iso, 'ADM1'),
                  result=Result.STRONG_MATCH),
            Query(where="name LIKE ? AND country = ?  AND f_code = ?",
                  args=(self.create_wildcard(lookup_target), place.country_iso, 'ADM1'),
                  result=Result.WILDCARD_MATCH),
            Query(where="name = ?  AND f_code = ?",
                  args=(lookup_target, 'ADM1'),
                  result=Result.SOUNDEX_MATCH)]

        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            place.admin1_id = row_list[0][Entry.ADM1]
            if place.country_iso == '':
                place.country_iso = row_list[0][Entry.ISO]

    def get_admin2_id(self, place: Loc):
        """Search for Admin1 entry"""
        lookup_target = place.admin2_name
        if len(lookup_target) == 0:
            return

        # Try each query until we find a match - each query gets less exact
        query_list = []
        if len(place.admin1_id) > 0:
            query_list.append(Query(where="name = ? AND country = ? AND admin1_id=? AND f_code=?",
                                    args=(lookup_target, place.country_iso, place.admin1_id, 'ADM2'),
                                    result=Result.STRONG_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? and admin1_id = ? AND f_code=?",
                                    args=(self.create_wildcard(lookup_target), place.country_iso, place.admin1_id, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? and admin1_id = ? AND f_code=?",
                                    args=(self.create_county_wildcard(lookup_target), place.country_iso, place.admin1_id, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))
        else:
            query_list.append(Query(where="name = ? AND country = ? AND f_code=?",
                                    args=(lookup_target, place.country_iso, 'ADM2'),
                                    result=Result.STRONG_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? AND f_code=?",
                                    args=(self.create_wildcard(lookup_target), place.country_iso, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))
            query_list.append(Query(where="name LIKE ? AND country = ? AND f_code=?",
                                    args=(self.create_county_wildcard(lookup_target), place.country_iso, 'ADM2'),
                                    result=Result.WILDCARD_MATCH))

        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            row = row_list[0]
            place.admin2_id = row[Entry.ADM2]

    def get_admin1_name(self, place: Loc) -> str:
        """Search for Admin1 entry"""
        lookup_target = place.admin1_id
        if len(lookup_target) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="admin1_id = ? AND country = ?  AND f_code = ? ",
                  args=(lookup_target, place.country_iso, 'ADM1'),
                  result=Result.STRONG_MATCH)
        ]
        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            row = row_list[0]
            place.admin1_name = row[Entry.NAME]
            return place.admin1_name
        else:
            return ''

    def get_admin1_name_direct(self, lookup_target, iso) -> str:
        """Search for Admin1 entry"""
        if len(lookup_target) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="admin1_id = ? AND country = ?  AND f_code = ? ",
                  args=(lookup_target, iso, 'ADM1'),
                  result=Result.STRONG_MATCH)
        ]
        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            row = row_list[0]
            return row[Entry.NAME]
        else:
            return ''

    def get_admin2_name_direct(self, admin1_id, admin2_id, iso) -> str:
        """Search for Admin2 entry"""
        lookup_target = admin2_id
        if len(lookup_target) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="admin2_id = ? AND country = ? AND admin1_id = ?",
                  args=(lookup_target, iso, admin1_id),
                  result=Result.STRONG_MATCH),
            Query(where="admin2_id = ? AND country = ?",
                  args=(lookup_target, iso),
                  result=Result.PARTIAL_MATCH)]

        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            row = row_list[0]
            return row[Entry.NAME]
        else:
            return ''

    def get_admin2_name(self, place: Loc) -> str:
        """Search for Admin1 entry"""
        lookup_target = place.admin2_id
        if len(lookup_target) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="admin2_id = ? AND country = ? AND admin1_id = ?",
                  args=(lookup_target, place.country_iso, place.admin1_id),
                  result=Result.STRONG_MATCH),
            Query(where="admin2_id = ? AND country = ?",
                  args=(lookup_target, place.country_iso),
                  result=Result.PARTIAL_MATCH)]

        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            row = row_list[0]
            place.admin2_name = row[Entry.NAME]
            # self.logger.debug(f'adm2 nm = {place.admin2_name}')
            return place.admin2_name
        else:
            return ''

    def lookup_geoid(self, place: Loc) -> None:
        """Search for GEOID"""
        result_place: Loc = Loc.Loc()

        query_list = [
            Query(where="geoid = ? ",
                  args=(place.target,),
                  result=Result.STRONG_MATCH)
        ]
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.geodata', query_list=query_list)
        if len(place.georow_list) == 0:
            place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)
        else:
            place.georow_list = place.georow_list[:1]
            place.result_type = GeoKeys.Result.STRONG_MATCH

        # Add search quality score to each entry
        for idx, rw in enumerate(place.georow_list):
            self.copy_georow_to_place(row=rw, place=result_place)
            update = list(rw)
            update.append(1)  # Extend list row and assign score
            result_place.prefix = ''
            res_nm = result_place.format_full_nm(None)
            # todo - move feature priority into scoring routine
            score = 0.0

            # Remove items in prefix that are in result
            tk_list = res_nm.split(",")
            for item in tk_list:
                place.prefix = re.sub(item.strip(' ').lower(), '', place.prefix)

            update[GeoKeys.Entry.SCORE] = int(score*100)
            place.georow_list[idx] = tuple(update)

    def lookup_main_dbid(self, place: Loc) -> None:
        """Search for DB ID"""
        query_list = [
            Query(where="id = ? ",
                  args=(place.target,),
                  result=Result.STRONG_MATCH)
        ]
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.geodata', query_list=query_list)

    def lookup_admin_dbid(self, place: Loc) -> None:
        """Search for DB ID"""
        query_list = [
            Query(where="id = ? ",
                  args=(place.target,),
                  result=Result.STRONG_MATCH)
        ]
        place.georow_list, place.result_type = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

    def get_country_name(self, iso: str) -> str:
        """ return country name for specified ISO code """
        if len(iso) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="country = ? AND f_code = ? ",
                  args=(iso, 'ADM0'),
                  result=Result.STRONG_MATCH)]

        row_list, res = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            res = row_list[0][Entry.NAME]
            if iso == 'us':
                res = 'United States'
        return res

    def get_country_iso(self, place: Loc) -> str:
        """ Return ISO code for specified country"""
        lookup_target = place.country_name
        if len(lookup_target) == 0:
            return ''

        # Try each query until we find a match - each query gets less exact
        query_list = [
            Query(where="name = ? AND f_code = ? ",
                  args=(lookup_target, 'ADM0'),
                  result=Result.STRONG_MATCH),
            # Query(where="name LIKE ?  AND f_code = ? ",
            #      args=(self.create_wildcard(lookup_target), 'ADM0'),
            #      result=Result.PARTIAL_MATCH)  #,
            # Query(where="sdx = ?  AND f_code = ? ",
            #      args=(GeoKeys.get_soundex (lookup_target), 'ADM0'),
            #      result=Result.PARTIAL_MATCH)
        ]

        row_list, result_code = self.db.process_query_list(from_tbl='main.admin', query_list=query_list)

        if len(row_list) > 0:
            res = row_list[0][Entry.ISO]
            if len(row_list) == 1:
                place.country_name = row_list[0][Entry.NAME]
        else:
            res = ''

        return res

    def copy_georow_to_place(self, row, place: Loc):
        # Copy data from DB row into Place
        # self.logger.debug(row)
        place.admin1_id = ''
        place.admin2_id = ''
        place.city1 = ''

        place.country_iso = str(row[Entry.ISO])
        place.country_name = str(self.get_country_name(row[Entry.ISO]))
        place.lat = row[Entry.LAT]
        place.lon = row[Entry.LON]
        place.feature = str(row[Entry.FEAT])
        place.geoid = str(row[Entry.ID])

        if place.feature == 'ADM0':
            self.place_type = Loc.PlaceType.COUNTRY
            pass
        elif place.feature == 'ADM1':
            place.admin1_id = row[Entry.ADM1]
            self.place_type = Loc.PlaceType.ADMIN1
        elif place.feature == 'ADM2':
            place.admin1_id = row[Entry.ADM1]
            place.admin2_id = row[Entry.ADM2]
            self.place_type = Loc.PlaceType.ADMIN2
        else:
            place.admin1_id = row[Entry.ADM1]
            place.admin2_id = row[Entry.ADM2]
            place.city1 = row[Entry.NAME]
            self.place_type = Loc.PlaceType.CITY

        place.admin1_name = str(self.get_admin1_name(place))
        place.admin2_name = str(self.get_admin2_name(place))
        if place.admin2_name is None:
            place.admin2_name = ''
        if place.admin1_name is None:
            place.admin1_name = ''

        place.city1 = str(place.city1)
        if place.city1 is None:
            place.city1 = ''

    def clear_geoname_data(self):
        # Delete all the geoname data
        for tbl in ['geodata', 'admin']:
            # noinspection SqlWithoutWhere
            self.db.delete_table(tbl)

    def create_geoid_index(self):
        # Create indices
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS geoid_idx ON geodata(geoid)')
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS admgeoid_idx ON admin(geoid)')

    def create_indices(self):
        # Create indices
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS name_idx ON geodata(name)')
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS country_idx ON geodata(country)')
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS admin1_idx ON geodata(admin1_id)')
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS sdx_idx ON geodata(sdx)')

        # self.db.create_index(create_table_sql='CREATE INDEX geoid_idx ON geodata(geoid)')

        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS adm_name_idx ON admin(name)')
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS adm_country_idx ON admin(country)')
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS adm_admin1_idx ON admin(admin1_id)')
        self.db.create_index(create_table_sql='CREATE INDEX IF NOT EXISTS adm_sdx_idx ON admin(sdx)')

    @staticmethod
    def make_georow(name: str, iso: str, adm1: str, adm2: str, lat: float, lon: float, feat: str, geoid: str, sdx: str) -> ():
        # Return a dummy geo-row
        res = (name, iso, adm1, adm2, lat, lon, feat, geoid, sdx)
        return res

    def get_row_count(self) -> int:
        return self.db.get_row_count()

    @staticmethod
    def create_wildcard(pattern):
        # Create SQL wildcard pattern (convert * to %).  Add % on end
        if '*' in pattern:
            return re.sub(r"\*", "%", pattern)
        else:
            return f'%{pattern}%'

    @staticmethod
    def create_county_wildcard(pattern):
        # Try pattern with 'shire' removed
        pattern = re.sub(r"shire", "", pattern)
        # Create SQL wildcard pattern (convert * to %).  Add % on end
        if '*' in pattern:
            return re.sub(r"\*", "%", pattern)
        else:
            return f'{pattern}'

    def close(self):
        self.logger.info('Closing Database')
        self.db.conn.close()

    def insert(self, geo_row: (), feat_code: str):
        # We split the data into 2  tables, 1) Admin: ADM0/ADM1/ADM2,  and 2) city data
        if feat_code == 'ADM1' or feat_code == 'ADM0' or feat_code == 'ADM2':
            sql = ''' INSERT OR IGNORE INTO admin(name,country, admin1_id,admin2_id,lat,lon,f_code, geoid, sdx)
                      VALUES(?,?,?,?,?,?,?,?,?) '''
            row_id = self.db.execute(sql, geo_row)
            # Add name to dictionary.  Used by AlternateNames for fast lookup during DB build
            self.geoid_admin_dict[geo_row[Entry.ID]] = row_id
        else:
            sql = ''' INSERT OR IGNORE INTO geodata(name, country, admin1_id, admin2_id, lat, lon, f_code, geoid, sdx)
                      VALUES(?,?,?,?,?,?,?,?,?) '''
            row_id = self.db.execute(sql, geo_row)
            # Add name to dictionary.  Used by AlternateNames for fast lookup during DB build
            self.geoid_main_dict[geo_row[Entry.ID]] = row_id

        return row_id

    def create_tables(self):
        # name, country, admin1_id, admin2_id, lat, lon, f_code, geoid
        sql_geodata_table = """CREATE TABLE IF NOT EXISTS geodata    (
                id           integer primary key autoincrement not null,
                name     text,
                country     text,
                admin1_id     text,
                admin2_id text,
                lat      text,
                lon       text,
                f_code      text,
                geoid      text,
                sdx     text
                                    );"""

        # name, country, admin1_id, admin2_id, lat, lon, f_code, geoid
        sql_admin_table = """CREATE TABLE IF NOT EXISTS admin    (
                id           integer primary key autoincrement not null,
                name     text,
                country     text,
                admin1_id     text,
                admin2_id text,
                lat      text,
                lon       text,
                f_code      text,
                geoid      text,
                sdx     text
                                    );"""

        for tbl in [sql_geodata_table, sql_admin_table]:
            self.db.create_table(tbl)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
import collections
import copy
import logging
import string as st
from operator import itemgetter

from geofinder import GeodataFiles, GeoKeys, Loc


class Geodata:
    """
    Provide a place lookup gazeteer based on datafiles from geonames.org read in by GeodataFiles.py

    The lookup returns whether a place exists and its lat/long.
    The data files must be downloaded from geonames.org
    First try a key lookup.  If that fails do a partial match
    """

    def __init__(self, directory_name: str, progress_bar):
        self.logger = logging.getLogger(__name__)
        self.status = "geoname file error"
        self.directory: st = directory_name
        self.progress_bar = progress_bar  # progress_bar
        self.geo_files = GeodataFiles.GeodataFiles(self.directory, progress_bar=self.progress_bar)  # , geo_district=self.geo_district)
        self.save_place = None
        self.last_iso = ''

    def update_prefix(self, place):
        temp_place = Loc.Loc()
        tokens = place.name.split(',')

        # Set all the prefix values in the georow_list
        for idx, rw in enumerate(place.georow_list):
            update = list(rw)

            # Put unused fields into prefix
            self.geo_files.geodb.copy_georow_to_place(rw, temp_place)
            temp_place.prefix = ''
            nm = GeoKeys.search_normalize(temp_place.format_full_nm(self.geo_files.output_replace_dct),place.country_iso)
            # self.logger.debug(f'NAME ={nm}')
            place.prefix = ''

            for num,fld in enumerate(tokens[:2]):
                item = GeoKeys.search_normalize(fld,place.country_iso)
                add_item = False
                # self.logger.debug(f'item={item} ')
                if num == 0 and item not in nm:
                    add_item = True

                if num == 1 and item not in nm and len(tokens) == 2:
                    # We only add the second token if there are only 2 tokens
                    add_item = True

                if '*' in item:
                    # Don't add as prefix if item is a wildcard search
                    add_item = False

                if add_item:
                    if len(place.prefix) > 0:
                        place.prefix += ' '
                    place.prefix += item.title()

            if len(place.prefix) > 0:
                place.prefix_commas = ', '
            update[GeoKeys.Entry.PREFIX] = place.prefix
            # self.logger.debug(f'PREFIX={place.prefix} ')

            place.georow_list[idx] = tuple(update)

    def lookup_by_type(self, place, result_list, typ, save_place):
        if typ == Loc.PlaceType.CITY:
            place.target = place.city1
            place.city1 = place.city1
            place.place_type = Loc.PlaceType.CITY
        elif typ == Loc.PlaceType.ADMIN2:
            place.target = place.admin2_name
            place.city1 = place.admin2_name
            place.admin2_name = ''
            place.place_type = Loc.PlaceType.CITY
        elif typ == Loc.PlaceType.ADMIN1:
            place.target = place.admin1_name
            place.city1 = place.admin1_name
            place.admin1_name = ''
            place.place_type = Loc.PlaceType.ADMIN1
        elif typ == Loc.PlaceType.PREFIX:
            place.target = place.prefix
            place.city1 = place.prefix
            place.prefix = ''
            place.place_type = Loc.PlaceType.CITY
        elif typ == Loc.PlaceType.COUNTRY:
            place.target = place.country_name
            place.place_type = Loc.PlaceType.COUNTRY

        self.geo_files.geodb.lookup_place(place=place)
        result_list.extend(place.georow_list)
        self.update_prefix(place=place)

        # Restore items
        place.city1 = save_place.city1
        place.admin2_name = save_place.admin2_name
        place.admin1_name = save_place.admin1_name
        place.prefix = save_place.prefix

    def find_location(self, location: str, place: Loc.Loc):
        """
        Find a location in the geoname dictionary.
        First parse the location into <prefix>, city, <district2>, district1, country.
        Then look it up in the place dictionary
        Update place with -- lat, lon, district, city, country_iso, result code
        """
        place.parse_place(place_name=location, geo_files=self.geo_files)

        flags = ResultFlags(limited=False, filtered=False)
        result_list = []

        #self.logger.debug(f'Find LOCATION City=[{place.city1}] Adm2=[{place.admin2_name}]\
        #Adm1=[{place.admin1_name}] Pref=[{place.prefix}] Cntry=[{place.country_name}] iso=[{place.country_iso}]  Type={place.place_type} ')

        self.save_place = copy.copy(place)

        if place.place_type == Loc.PlaceType.ADVANCED_SEARCH:
            # Lookup location with advanced search params
            self.logger.debug('Advanced Search')
            self.lookup_by_type(place, result_list, place.place_type, self.save_place)
            return

        self.country_is_valid(place)

        # The country in this entry is not supported (not loaded into DB)
        if place.result_type == GeoKeys.Result.NOT_SUPPORTED:
            self.process_result(place=place, targ_name=place.target, flags=flags)
            return

        # Try standard lookup based on simple parse:  city, county, state/province, country
        self.geo_files.geodb.lookup_place(place=place)
        result_list.extend(place.georow_list)
        self.update_prefix(place=place)

        # Simple parse can be wrong, so also try prefix token  as city and admin2 token as city
        for ty in [Loc.PlaceType.PREFIX, Loc.PlaceType.ADMIN2]:
            self.lookup_by_type(place, result_list, ty, self.save_place)

        #self.logger.debug(result_list)

        #  Move result list into place georow list
        place.georow_list.clear()
        place.georow_list.extend(result_list)

        if len(place.georow_list) > 0:
            # Build list - sort and remove duplicates
            #self.logger.debug(f'Match {place.georow_list}')
            self.process_result(place=place, targ_name=place.target, flags=flags)
            flags = self.build_result_list(place)

        if len(place.georow_list) == 0:
            # NO MATCH
            self.logger.debug(f'Not found.')
            # place = self.save_place
            if place.result_type != GeoKeys.Result.NO_COUNTRY and place.result_type != GeoKeys.Result.NOT_SUPPORTED:
                place.result_type = GeoKeys.Result.NO_MATCH
        elif len(place.georow_list) > 1:
            self.logger.debug(f'Success!  {len(place.georow_list)} matches')
            place.result_type = GeoKeys.Result.MULTIPLE_MATCHES

        # Process the results
        self.process_result(place=place, targ_name=place.target, flags=flags)
        self.logger.debug(f'Status={place.status}')

    def lookup_geoid(self, place):
        flags = ResultFlags(limited=False, filtered=False)
        self.geo_files.geodb.lookup_geoid(place)
        self.update_prefix(place=place)
        self.process_result(place,'', flags=flags)

    def search_city(self, place):
        place.target = place.city1
        place.prefix = f' {place.admin2_name.title()}'
        self.logger.debug(f' Try city [{place.target}] as city')
        self.geo_files.geodb.lookup_place(place=place)
        self.update_prefix(place=place)

    def search_admin2(self, place):
        place.target = place.admin2_name
        place.prefix = f' {place.city1.title()}'
        self.logger.debug(f'  Try admin2  [{place.target}] as city')
        self.geo_files.geodb.lookup_place(place=place)
        self.update_prefix(place=place)

    def search_admin1(self, place):
        place.target = place.admin1_name
        # place.prefix = f' {place.city1.title()} {place.admin2_name.title()}'
        self.logger.debug(f'  Try admin1  [{place.target}] as city')
        self.geo_files.geodb.lookup_place(place=place)
        self.update_prefix(place=place)

    def process_result(self, place: Loc.Loc, targ_name, flags) -> None:
        # Copy geodata to place record and Put together status text
        #self.logger.debug(f'**PROCESS RESULT:  Res={place.result_type}  Targ={place.target} Georow_list={place.georow_list}')
        if place.result_type == GeoKeys.Result.NOT_SUPPORTED:
            place.place_type = Loc.PlaceType.COUNTRY

        if place.result_type in GeoKeys.successful_match and len(place.georow_list) > 0:
            self.geo_files.geodb.copy_georow_to_place(row=place.georow_list[0], place=place)
            place.format_full_nm(self.geo_files.output_replace_dct)
        elif len(place.georow_list) > 0:
            self.logger.debug(f'***RESULT={place.result_type} Setting to Partial')
            place.result_type = GeoKeys.Result.PARTIAL_MATCH

        place.prefix = place.prefix.strip(' ')
        self.set_place_type_text(place=place)
        place.status = f'{place.result_type_text}  {result_text_list.get(place.result_type)} '
        if flags.limited:
            place.status += ' First 300 matches shown...'

        if flags.filtered:
            place.status = f'{place.result_type_text}  {result_text_list.get(place.result_type)} '
            place.status += ' ***VERIFY EVENT DATE***'
            place.result_type = GeoKeys.Result.PARTIAL_MATCH

        self.update_prefix(place=place)

    def find_first_match(self, location: st, place: Loc.Loc):
        """
        Find the first match for this location in the geoname dictionary.
        First parse the location into <prefix>, city, <district2>, district1, country.
        Then look it up in the place db
        Update place with -- lat, lon, district, city, country_iso, result code
        """
        place.parse_place(place_name=location, geo_files=self.geo_files)
        place.country_name = self.geo_files.geodb.get_country_name(place.country_iso)

        place.country_iso = place.country_iso

        # Lookup location
        self.geo_files.geodb.lookup_place(place=place)
        self.update_prefix(place=place)

        # Clear to a single entry
        if len(place.georow_list) > 1:
            row = copy.copy(place.georow_list[0])
            place.georow_list.clear()
            place.georow_list.append(row)
            place.result_type = GeoKeys.Result.STRONG_MATCH

        self.process_result(place=place, targ_name=place.target, flags=ResultFlags(limited=False, filtered=False))

    def find_geoid(self, geoid: str, place: Loc.Loc):
        place.target = geoid
        place.georow_list.clear()
        self.geo_files.geodb.lookup_geoid(place=place)
        if len(place.georow_list) > 0:
            # Copy geo row to Place
            self.geo_files.geodb.copy_georow_to_place(row=place.georow_list[0], place=place)
            place.name = place.format_full_nm(None)
            place.result_type = GeoKeys.Result.STRONG_MATCH
        else:
            place.result_type = GeoKeys.Result.NO_MATCH

    def set_place_type_text(self, place: Loc.Loc):
        if place.result_type == GeoKeys.Result.NO_COUNTRY:
            place.result_type_text = 'Country'
        elif place.place_type == Loc.PlaceType.COUNTRY:
            place.result_type_text = 'Country'
        elif place.place_type == Loc.PlaceType.ADMIN1:
            place.result_type_text = self.get_district1_type(place.country_iso)
        elif place.place_type == Loc.PlaceType.ADMIN2:
            place.result_type_text = 'County'
        elif place.place_type == Loc.PlaceType.CITY:
            place.result_type_text = self.get_type_name(place.feature)
        elif place.place_type == Loc.PlaceType.PREFIX:
            place.result_type_text = 'Place'
        """
        if place.place_type == Loc.PlaceType.CITY:
            place.result_type_text = '' #GeoKeys.type_names.get(place.feature)
            if place.result_type_text is None:
                place.result_type_text = ' '
        elif place.place_type == Loc.PlaceType.ADMIN1:
            place.result_type_text = self.get_district1_type(place.country_iso)
        else:
            place.result_type_text = Loc.place_type_name_dict[place.place_type]
        """

    def set_last_iso(self, iso):
        self.last_iso = iso

    def read(self) -> bool:
        """ Read in geo name files which contain place names and their lat/lon.
            Return True if error
        """
        err = self.geo_files.read()
        if err:
            return True

    def read_geonames(self):
        self.progress("Reading Geoname files...", 70)
        return self.geo_files.read_geoname()

    def progress(self, msg: st, percent: int):
        if self.progress_bar is not None:
            self.progress_bar.update_progress(percent, msg)
        else:
            self.logger.debug(msg)

    def country_is_valid(self, place: Loc) -> bool:
        # See if COUNTRY is present and is in the supported country list
        if place.country_iso == '':
            place.result_type = GeoKeys.Result.NO_COUNTRY
            is_valid = False
        elif place.country_iso not in self.geo_files.supported_countries_dct:
            self.logger.debug(f'[{place.country_iso}] not supported')
            place.result_type = GeoKeys.Result.NOT_SUPPORTED
            place.place_type = Loc.PlaceType.COUNTRY
            place.target = place.country_name
            is_valid = False
        else:
            is_valid = True

        return is_valid

    def validate_year_for_location(self, event_year: int, iso: str, admin1: str, padding: int) -> bool:
        # See if this location name was valid at the time of the event
        # Try looking up start year by state/province
        event_year += padding
        start_year = admin1_name_start_year.get(f'{iso}.{admin1.lower()}')
        if start_year is None:
            # Try looking up start year by country
            start_year = country_name_start_year.get(iso)
        if start_year is None:
            start_year = -1

        if event_year < start_year and event_year != 0 + padding:
            # self.logger.debug(f'Val year:  loc year={start_year}  event yr={event_year} loc={admin1},{iso}')
            return False
        else:
            return True

    def build_result_list(self, place):
        # Create a sorted version of result_list without any dupes
        # Add flag if we hit the lookup limit
        # Discard location names that didnt exist at time of event and add to result flag
        date_filtered = False  # Flag to indicate whether we dropped locations due to event date
        event_year = place.event_year

        if len(place.georow_list) > 299:
            limited_flag = True
        else:
            limited_flag = False

        # sort list by State/Province id, and County id
        list_copy = sorted(place.georow_list, key=itemgetter(GeoKeys.Entry.ADM1, GeoKeys.Entry.LON))
        place.georow_list.clear()
        distance_cutoff = 0.5  # Value to determine if two lat/longs are similar

        # Create a dummy 'previous' row so first comparison works
        prev_geo_row = self.geo_files.geodb.make_georow(name='q', iso='q', adm1='q', adm2='q', lat=900, lon=900, feat='q', geoid='q', sdx='q')
        idx = 0

        # Create new list without dupes (adjacent items with same name and same lat/lon)
        # Find if two items with same name are similar lat/lon (within Box Distance of 0.5 degrees)
        #self.logger.debug('===== BUILD RESULT =====')
        for geo_row in list_copy:
            if self.validate_year_for_location(event_year, geo_row[GeoKeys.Entry.ISO], geo_row[GeoKeys.Entry.ADM1], 80) is False:
                # Skip location if location name  didnt exist at the time of event WITH 80 years padding
                continue

            if self.validate_year_for_location(event_year, geo_row[GeoKeys.Entry.ISO], geo_row[GeoKeys.Entry.ADM1], 0) is False:
                # Flag if location name  didnt exist at the time of event
                date_filtered = True

            new_row = list(geo_row)
            geo_row = tuple(new_row)
            self.logger.debug(f'{geo_row[GeoKeys.Entry.NAME]},{geo_row[GeoKeys.Entry.FEAT]} '
                              f'{geo_row[GeoKeys.Entry.SCORE]:.1f} {geo_row[GeoKeys.Entry.ADM2]}, '
                              f'{geo_row[GeoKeys.Entry.ADM1]} {geo_row[GeoKeys.Entry.ISO]}')

            if geo_row[GeoKeys.Entry.NAME] != prev_geo_row[GeoKeys.Entry.NAME]:
                # Name is different.  Add previous item
                place.georow_list.append(geo_row)
                idx += 1
            elif abs(float(prev_geo_row[GeoKeys.Entry.LAT]) - float(geo_row[GeoKeys.Entry.LAT])) + \
                    abs(float(prev_geo_row[GeoKeys.Entry.LON]) - float(geo_row[GeoKeys.Entry.LON])) > distance_cutoff:
                # Lat/lon is different from previous item. Add this one
                place.georow_list.append(geo_row)
                idx += 1
            elif self.get_priority(geo_row[GeoKeys.Entry.FEAT]) < self.get_priority(prev_geo_row[GeoKeys.Entry.FEAT]):
                # Same Lat/lon but this has higher feature priority so replace previous entry
                place.georow_list[idx - 1] = geo_row

            prev_geo_row = geo_row

        min_score = 9999
        new_list = sorted(place.georow_list, key=itemgetter(GeoKeys.Entry.SCORE, GeoKeys.Entry.ADM1, GeoKeys.Entry.ADM2))
        place.georow_list.clear()

        for rw, geo_row in enumerate(new_list):
            # if rw==0:
            #   place.prefix = geo_row[GeoKeys.Entry.PREFIX].title()
            #  place.prefix = place.prefix.strip(' ')
            score = geo_row[GeoKeys.Entry.SCORE]
            if score < min_score:
                min_score = score
            #self.logger.debug(f'Score {score:.2f}  {geo_row[GeoKeys.Entry.NAME]}, {geo_row[GeoKeys.Entry.ADM2]}, {geo_row[GeoKeys.Entry.ADM1]}')
            if score > min_score + 15 or score > 88:
                break
            place.georow_list.append(geo_row)

        if min_score < 20 and len(place.georow_list) == 1:
            place.result_type = GeoKeys.Result.STRONG_MATCH

        return ResultFlags(limited=limited_flag, filtered=date_filtered)

    @staticmethod
    def get_priority(feature):
        f_prior = feature_priority.get(feature)
        if f_prior is None:
            f_prior = 1

        return (22.0 - float(f_prior) )/ 6.0

    @staticmethod
    def get_type_name(feature):
        nm = type_name.get(feature)
        if nm is None:
            nm = ''
        return nm

    @staticmethod
    def get_district1_type(iso) -> str:
        # Return the local country term for Admin1 district
        if iso in ["al"]:
            return "County"
        elif iso in ["us", "at", "bm", "br", "de"]:
            return "State"
        elif iso in ["ac", "an", 'ao', 'bb', 'bd']:
            return "Parish"
        elif iso in ["ae"]:
            return "Emirate"
        elif iso in ["bc", "bf", "bh", "bl", "bn"]:
            return "District"
        elif iso in ["gb"]:
            return "Country"
        else:
            return "Province"


default = ["ADM1", "ADM2", "ADM3", "ADM4", "ADMF", "CH", "CSTL", "CMTY", "EST ", "HSP",
           "HSTS", "ISL", "MSQE", "MSTY", "MT", "MUS", "PAL", "PPL", "PPLA", "PPLA2", "PPLA3", "PPLA4",
           "PPLC", "PPLG", "PPLH", "PPLL", "PPLQ", "PPLX", "PRK", "PRN", "PRSH", "RUIN", "RLG", "STG", "SQR", "SYG", "VAL"]

# If there are 2 identical entries, we only add the one with higher feature priority.  Highest value is for large city or capital
feature_priority = {'PP1M':22, 'PP1K':19, 'ADM1': 20, 'PPLA': 20, 'PPL': 18, 'PPLA2': 19, 'PPLA3': 17, 'PPLA4': 16, 'PPLC': 20, 'PPLG': 15,
'ADM2': 19, 'MILB': 13,'NVB': 12, 'PPLF': 11, 'DEFAULT': 3, 'ADM0': 10, 'PPLL': 6, 'PPLQ': 5, 'PPLR': 4, 'PPLS': 3, 'PPLW': 3, 'PPLX': 3, 'BTL': 3,
                    'STLMT': 1, 'CMTY': 4, 'VAL': 1, 'CH': 4, 'MSQE': 4}

result_text_list = {
    GeoKeys.Result.STRONG_MATCH: 'Matched! Click Save to accept:',
    GeoKeys.Result.MULTIPLE_MATCHES: ' Multiple matches.  Select one and click Verify or Double-Click',
    GeoKeys.Result.NO_MATCH: 'Not found.  Edit and click Verify.',
    GeoKeys.Result.NOT_SUPPORTED: ' Country is not supported. Skip or Add Country in Config',
    GeoKeys.Result.NO_COUNTRY: 'No Country found.',
    GeoKeys.Result.PARTIAL_MATCH: 'Partial match.  Click Save to accept:',
    GeoKeys.Result.DELETE: 'Empty.  Click Save to delete entry.',
    GeoKeys.Result.WILDCARD_MATCH: 'Wildcard match. Click Save to accept:',
    GeoKeys.Result.SOUNDEX_MATCH: 'Soundex match. Click Save to accept:',
}

ResultFlags = collections.namedtuple('ResultFlags', 'limited filtered')

# Starting year this country name was valid
country_name_start_year = {
    'cu': -1,
}

# Starting year this state/province modern names were valid
# https://en.wikipedia.org/wiki/List_of_North_American_settlements_by_year_of_foundation
admin1_name_start_year = {
    'us.al': 1711,
    'us.ak': 1774,
    'us.az': 1775,
    'us.ar': 1686,
    'us.ca': 1769,
    'us.co': 1871,
    'us.ct': 1633,
    'us.de': 1638,
    'us.dc': 1650,
    'us.fl': 1565,
    'us.ga': 1566,
    'us.hi': -1,
    'us.id': 1862,
    'us.il': 1703,
    'us.in': 1715,
    'us.ia': 1785,
    'us.ks': 1870,
    'us.ky': 1775,
    'us.la': 1699,
    'us.me': 1604,
    'us.md': 1633,
    'us.ma': 1620,
    'us.mi': 1784,
    'us.mn': 1820,
    'us.ms': 1699,
    'us.mo': 1765,
    'us.mt': 1877,
    'us.ne': 1854,
    'us.nv': 1905,
    'us.nh': 1638,
    'us.nj': 1624,
    'us.nm': 1598,
    'us.ny': 1614,
    'us.nc': 1653,
    'us.nd': 1871,
    'us.oh': 1785,
    'us.ok': 1889,
    'us.or': 1811,
    'us.pa': 1682,
    'us.ri': 1636,
    'us.sc': 1663,
    'us.sd': 1865,
    'us.tn': 1739,
    'us.tx': 1685,
    'us.ut': 1847,
    'us.vt': 1650,
    'us.va': 1607,
    'us.wa': 1825,
    'us.wv': 1788,
    'us.wi': 1685,
    'us.wy': 1867,
    'ca.01': 1795,
    'ca.02': 1789,
    'ca.03': 1733,
    'ca.04': 1766,
    'ca.05': 1583,
    'ca.07': 1604,
    'ca.08': 1673,
    'ca.09': 1764,
    'ca.10': 1541,
    'ca.11': 1862,
    'ca.12': 1700,
    'ca.13': 1700,
    'ca.14': 1700
}

type_name = {"ADM0":'Country', "ADM1":'City', "ADM2":'City', "ADM3":'City', "ADM4":'City', "ADMF":'City',
             "CH":'Church', "CSTL":'Castle', "CMTY":'Cemetery', "EST":'Estate', "HSP":'Hospital',
           "HSTS":'Historic', "ISL":'Island', "MSQE":'Mosque', "MSTY":'Monastery', "MT":'Mountain', "MUS":'Museum', "PAL":'Palace',
             "PPL":'City', "PPLA":'City', "PPLA2":'City', "PPLA3":'City', "PPLA4":'City',
           "PPLC":'City', "PPLG":'City', "PPLH":'City', "PPLL":'Village', "PPLQ":'City', "PPLX":'City',
            "PRK":'Park', "PRN":'Prison', "PRSH":'Parish', "RUIN":'Ruin',
            "RLG":'Religious', "STG":'', "SQR":'Square', "SYG":'Synagogue', "VAL":'Valley', "PP1M":'City', "PP1K":'City'}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

from tkinter import ttk
from tkinter.ttk import Progressbar


class Progress:
    """ Create a progress bar for loading large files or other long operations """

    def __init__(self, window, bar_color, trough_color, status):
        self.window = window
        self.shutdown_requested = False

        style: ttk.Style = ttk.Style()
        style.configure("bar.Horizontal.TProgressbar",
                        troughcolor=trough_color, bordercolor='white',
                        background=bar_color, lightcolor=bar_color, darkcolor=bar_color)

        self.bar: Progressbar = Progressbar(window, length=400, maximum=100, style="bar.Horizontal.TProgressbar")
        self.bar['value'] = 0
        self.lable = status
        self.startup: bool = True

        # Call appropriate update loop depending on whether we have entered main window loop
        if self.startup:
            self.window.update()  # Still in startup mode.  Use stronger window.update
        else:
            self.window.update_idletasks()

    def update_progress(self, progress: int, stage: str):
        self.bar['value'] = progress
        self.lable.set_text(stage)
        if self.startup:
            self.window.update()
        else:
            self.window.update_idletasks()

    def destroy(self):
        self.bar.destroy()
        self.window.update_idletasks()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
import os
import re
from typing import Match, Union

from geofinder import Progress
from geofinder.AncestryFile import AncestryFile
from geofinder.CachedDictionary import CachedDictionary

PLACE_TOTAL_KEY = 'PLACE_TOTAL'


class Gedcom(AncestryFile):
    """
    Class for Gedcom file handler - based on AncestryFile handler
    Basic routines to Read/Parse and Write GEDCOM ancestry files focused on place entries.
    Scan - Read through  file, find Place entry.
    Write out all other entries as-is if out_path is not None
    """

    def __init__(self, in_path: str, out_suffix: str, cache_d, progress: Union[None, Progress.Progress],geodata):
        super().__init__(in_path, out_suffix, cache_d, progress, geodata)

        # Sections of a GEDCOM line - Level, label, tag, value
        self.level: int = 0
        self.label: str = ""

        # Build dictionary of name/id pairs and write to pickle cache file.  If pickle file already there, just read it.
        # When we display a location, we use this to display the name the event is tied to
        parts = os.path.split(in_path)
        filename = parts[1] + '.pkl'
        self.person_cd = CachedDictionary(cache_d, filename)

        # Try to read pickle file of IDs for this GEDCOM file
        err = self.person_cd.read()
        if err:
            # File is not there.  Build it - it is a dictionary of GED Name_IDs to Names
            self.build_person_dictionary()
        else:
            # Get Place count from person dictionary
            self.place_total = self.person_cd.dict.get(PLACE_TOTAL_KEY)
            self.logger.debug(f'Place Total ={self.place_total}')

    def parse_line(self, line: str):
        # Called by read_and_parse_line for each line in file.  Parse line
        # and returns each place entry in self.value with self.tag set to PLAC

        # Gedcom file regex:          Digits for level,   @  for label,   text for tag,   text for value
        regex = re.compile(r"^(?P<level>\d+)\s+(?P<label>@\S+@)?\s*(?P<tag>\S+)\s+(?P<value>.*)")

        """ Parse GEDCOM line to Level, Label (if present), Tag, Value. """
        matches: Match = regex.match(line)

        if matches is not None:
            self.tag = matches.group('tag')  # GEDCOM tag
            self.level = int(matches.group('level'))  # GEDCOM level
            self.value = matches.group('value')  # GEDCOM value for command
            self.value = self.value.rstrip("\\")
            self.label = matches.group('label')  # GEDCOM label
        else:
            # Could not parse
            self.tag = ""
            self.value = ""
            self.level = 99
            self.label = ''

        # update progress bar
        self.percent_complete = int(self.infile.tell() * 100 / self.filesize)
        if self.line_num % 1000 == 1:
            self.progress(f"Scanning ", self.percent_complete)

        return self.id

    def write_updated(self, txt: str, place):
        """ Write out a place line with updated value.  Put together the pieces:  level, Label, tag, value """
        if self.outfile is not None:
            if self.label is not None:
                res = f"{self.level} {self.label} {self.tag} {txt.strip(', ')}\n"
            else:
                res = f"{self.level} {self.tag} {txt.strip(', ')}\n"

            self.outfile.write(res)

    def write_asis(self, entry):
        """ Write out a place line as-is.  Put together the pieces:  level, Label, tag, value """
        if self.outfile is not None:
            if self.label is not None:
                res = f"{self.level} {self.label} {self.tag} {self.value}\n"
            else:
                res = f"{self.level} {self.tag} {self.value}\n"

            self.outfile.write(res)

    def write_lat_lon(self, lat: float, lon: float):
        """ Write out a GEDCOM PLACE MAP entry with latitude and longitude. """
        if self.output_latlon is False:
            return

        if self.outfile is not None:
            map_level: int = self.level + 1
            lati_level: int = self.level + 2

            # Output Lat / Long
            if lon != float('NaN'):
                #  If there is already a MAP LATI LONG entry, eat it without output
                line: str = self.peak_next_line()
                self.parse_line(line)

                if self.tag == "MAP":
                    # Read this MAP command and do nothing with it
                    self.infile.readline()

                    # Check for LATI line
                    line = self.peak_next_line()
                    self.parse_line(line)
                    if self.tag == "LATI" or self.tag == "LONG":
                        # Read this LATI command and do nothing with it
                        self.infile.readline()

                    # Check for LONG line
                    line = self.peak_next_line()
                    self.parse_line(line)
                    if self.tag == "LATI" or self.tag == "LONG":
                        # Read this LONG command and do nothing with it
                        self.infile.readline()

                # Write out MAP Latitude/Longitude section
                self.outfile.write(f"{str(map_level)} MAP\n")
                self.outfile.write(f"{str(lati_level)} LATI {lat}\n")
                self.outfile.write(f"{str(lati_level)} LONG {lon}\n")

    def collect_event_details(self):
        """ Collect details for event - last name, event date, and tag in GEDCOM file."""

        # Text names for event tags
        event_names = {'DEAT': 'Death', 'CHR': 'Christening', 'BURI': 'Burial', 'BIRT': 'Birth',
                       'CENS': 'Census', 'MARR': 'Marriage', 'RESI': 'Residence', 'IMMI': 'Immigration', 'EMMI': 'Emmigration',
                       'OCCU': 'Occupation'}

        # Level of 0 indicates a new record - reset values
        if self.level == 0:
            self.id = ' '
            self.name = ''
            self.clear_date()

            if self.tag == 'INDI' or self.tag == 'FAM':
                if self.label is not None:
                    self.id = self.label
                    self.name = self.label
                    self.event_name = self.tag
                else:
                    self.logger.info(f'NO label line {self.line_num} tag {self.tag}')
        elif self.level == 1:
            self.clear_date()
            if self.tag == 'NAME':
                self.name = self.value
            if self.tag == 'HUSB':
                # We cheat on the Family tag and just use the Husbands name
                self.name = self.value
            # Store name of events that have Locations
            if self.tag in event_names:
                self.event_name = event_names[self.tag]
                self.clear_date()
            elif self.tag == 'TYPE':
                self.event_name = self.value
            elif self.tag != 'DATE' and self.tag != 'PLAC':
                self.event_name = self.tag
        elif self.level == 2:
            if self.tag == 'DATE':
                self.set_date(self.value)
            elif self.tag == 'TYPE':
                self.event_name = self.value

    def set_date(self, date: str):
        """ Set Date and Parse string for date/year and set Gedcom year of event """
        # Only supports simple date and ABT date, ignores date ranges
        self.date = date
        self.event_year = 0
        self.abt_flag = False  # Flag to indicate that this is an "ABOUT" date

        # Support DATE and ABT DATE of form <DD> <MMM> YYYY (GEDCOM format) with no validation
        reg = r'^\s*(ABT\s+)?([1-3]?[0-9]{1}\s+)?((JAN|FEB|MAR|APR|MAY|JUN|JUL|AUG|SEP|OCT|NOV|DEC)\s+)?(\d{3,4})'

        m = re.search(reg, date)
        if m:
            abt = (m.group(1))
            # day = (m.group(2))
            # mon = (m.group(3))
            # group4 not used
            year = (m.group(5))

            if year is not None:
                self.event_year = int(year)
            if abt is not None:
                self.abt_flag = True

    def clear_date(self):
        self.event_year = 0
        self.date = ''

    def build_person_dictionary(self):
        """
        Read gedcom and extract Person names
        This is used to do lookup from ID to name
        """
        while True:
            line, err = self.read_and_parse_line()
            if err:
                break  # END OF FILE

            if self.tag == 'NAME' or self.tag == 'HUSB':
                # self.logger.debug(f'ky=[{self.id}] val=[{self.value}]')
                if self.id != self.value:
                    self.person_cd.dict[self.id] = self.value

            if self.tag == 'PLAC':
                self.place_total += 1

        # Save place total
        self.person_cd.dict[PLACE_TOTAL_KEY] = self.place_total
        self.logger.debug(f'Place Total ={self.place_total}')

        # Write out cached dictionary
        self.person_cd.write()

        # Done.  Reset file back to start
        self.infile.seek(0)
        self.line_num = 0
        self.logger.debug('build ged done')
        self.build = True

    def get_name(self, nam: str, depth: int = 0) -> str:
        # Get name of person we are currently on
        nm = self.person_cd.dict.get(nam)
        if nm is not None:
            if nm[0] == '@' and depth < 4:
                # Recursively call to get through the '@' indirect values.  make sure we don't go too deep
                nm = self.get_name(nm, depth + 1)
        else:
            nm = self.name

        self.logger.debug(f'{depth}) ky={self.id} {nm}: [{self.event_name}] [{self.date}]')

        return nm.replace('/', '')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

import logging
from typing import Dict

import phonetics

from geofinder import GeoDB, GeoKeys


class CnRow:
    ISO = 0
    ISO3 = 1
    NUM = 2
    LAT = 3
    LON = 4


class Country:
    def __init__(self, progress, geodb, lang_list):
        self.logger = logging.getLogger(__name__)
        self.geodb = geodb
        self.progress = progress
        self.country_dict: Dict[str, str] = {}  # Dictionary of Country Name to country iso code
        self.iso_dict: Dict[str, str] = {}  # Reverse dictionary - Country ISO code to country name
        self.lang_list = lang_list

    def read(self) -> bool:
        """
           Read in list of country names and ISO codes
        """
        if self.progress is not None:
            self.progress.update_progress(100, "Read ISO countries...")

        # list of all countries and their ISO codes
        # This also includes some common aliases
        self.geodb.db.begin()

        self.logger.debug(self.lang_list)

        #  Add country names to DB
        for ky, row in country_dict.items():
            # Localize country names to specified
            for lang in self.lang_list:
                # If we have a translation table for this language, then apply it
                if trans_table.get(lang):
                    tbl = trans_table.get(lang)
                    # Look up the country translation
                    if tbl.get(ky):
                        ky = tbl.get(ky)
                    break  # Apply first translation in list

            # Create Geo_row
            # ('paris', 'fr', '07', '012', '12.345', '45.123', 'PPL')
            geo_row = [None] * GeoDB.Entry.MAX
            geo_row[GeoDB.Entry.NAME] = GeoKeys.normalize(ky)
            sdx = phonetics.dmetaphone(geo_row[GeoDB.Entry.NAME])
            geo_row[GeoDB.Entry.SDX] = sdx[0]

            geo_row[GeoDB.Entry.ISO] = row[CnRow.ISO].lower()
            geo_row[GeoDB.Entry.ADM1] = ''
            geo_row[GeoDB.Entry.ADM2] = ''
            geo_row[GeoDB.Entry.LAT] = row[CnRow.LAT]
            geo_row[GeoDB.Entry.LON] = row[CnRow.LON]
            geo_row[GeoDB.Entry.FEAT] = 'ADM0'
            geo_row[GeoDB.Entry.ID] = row[CnRow.ISO].lower()

            self.geodb.insert(geo_row=geo_row, feat_code='ADM0')

        self.geodb.db.commit()
        return False


country_dict = {
    'Afghanistan': ('AF', 'AFG', '4', '33', '65'),
    'Albania': ('AL', 'ALB', '8', '41', '20'),
    'Algeria': ('DZ', 'DZA', '12', '28', '3'),
    'American Samoa': ('AS', 'ASM', '16', '-14.333', '-170'),
    'Andorra': ('AD', 'AND', '20', '42.5', '1.6'),
    'Angola': ('AO', 'AGO', '24', '-12.5', '18.5'),
    'Anguilla': ('AI', 'AIA', '660', '18.25', '-63.166'),
    'Antarctica': ('AQ', 'ATA', '10', '-90', '0'),
    'Antigua and Barbuda': ('AG', 'ATG', '28', '17.05', '-61.8'),
    'Argentina': ('AR', 'ARG', '32', '-34', '-64'),
    'Armenia': ('AM', 'ARM', '51', '40', '45'),
    'Aruba': ('AW', 'ABW', '533', '12.5', '-69.966'),
    'Australia': ('AU', 'AUS', '36', '-27', '133'),
    'Austria': ('AT', 'AUT', '40', '47.333', '13.333'),
    'Azerbaijan': ('AZ', 'AZE', '31', '40.5', '47.5'),
    'Bahamas': ('BS', 'BHS', '44', '24.25', '-76'),
    'Bahrain': ('BH', 'BHR', '48', '26', '50.55'),
    'Bangladesh': ('BD', 'BGD', '50', '24', '90'),
    'Barbados': ('BB', 'BRB', '52', '13.166', '-59.533'),
    'Belarus': ('BY', 'BLR', '112', '53', '28'),
    'Belgium': ('BE', 'BEL', '56', '50.833', '4'),
    'Belize': ('BZ', 'BLZ', '84', '17.25', '-88.75'),
    'Benin': ('BJ', 'BEN', '204', '9.5', '2.25'),
    'Bermuda': ('BM', 'BMU', '60', '32.333', '-64.75'),
    'Bhutan': ('BT', 'BTN', '64', '27.5', '90.5'),
    'Bolivia, Plurinational State of': ('BO', 'BOL', '68', '-17', '-65'),
    'Bolivia': ('BO', 'BOL', '68', '-17', '-65'),
    'Bosnia and Herzegovina': ('BA', 'BIH', '70', '44', '18'),
    'Botswana': ('BW', 'BWA', '72', '-22', '24'),
    'Bouvet Island': ('BV', 'BVT', '74', '-54.433', '3.4'),
    'Brazil': ('BR', 'BRA', '76', '-10', '-55'),
    'British Indian Ocean Territory': ('IO', 'IOT', '86', '-6', '71.5'),
    'Brunei Darussalam': ('BN', 'BRN', '96', '4.5', '114.666'),
    'Brunei': ('BN', 'BRN', '96', '4.5', '114.666'),
    'Bulgaria': ('BG', 'BGR', '100', '43', '25'),
    'Burkina Faso': ('BF', 'BFA', '854', '13', '-2'),
    'Burundi': ('BI', 'BDI', '108', '-3.5', '30'),
    'Cambodia': ('KH', 'KHM', '116', '13', '105'),
    'Cameroon': ('CM', 'CMR', '120', '6', '12'),
    'Canada': ('CA', 'CAN', '124', '60', '-95'),
    'Cape Verde': ('CV', 'CPV', '132', '16', '-24'),
    'Cayman Islands': ('KY', 'CYM', '136', '19.5', '-80.5'),
    'Central African Republic': ('CF', 'CAF', '140', '7', '21'),
    'Chad': ('TD', 'TCD', '148', '15', '19'),
    'Chile': ('CL', 'CHL', '152', '-30', '-71'),
    'China': ('CN', 'CHN', '156', '35', '105'),
    'Christmas Island': ('CX', 'CXR', '162', '-10.5', '105.666'),
    'Cocos Keeling) Islands': ('CC', 'CCK', '166', '-12.5', '96.833'),
    'Colombia': ('CO', 'COL', '170', '4', '-72'),
    'Comoros': ('KM', 'COM', '174', '-12.166', '44.25'),
    'Congo': ('CG', 'COG', '178', '-1', '15'),
    'Congo, the Democratic Republic of the': ('CD', 'COD', '180', '0', '25'),
    'Cook Islands': ('CK', 'COK', '184', '-21.233', '-159.766'),
    'Costa Rica': ('CR', 'CRI', '188', '10', '-84'),
    'Côte d Ivoire': ('CI', 'CIV', '384', '8', ' - 5'),
    'Ivory Coast': ('CI', 'CIV', '384', '8', '-5'),
    'Croatia': ('HR', 'HRV', '191', '45.166', '15.5'),
    'Cuba': ('CU', 'CUB', '192', '21.5', '-80'),
    'Cyprus': ('CY', 'CYP', '196', '35', '33'),
    'Czech Republic': ('CZ', 'CZE', '203', '49.75', '15.5'),
    'Denmark': ('DK', 'DNK', '208', '56', '10'),
    'Djibouti': ('DJ', 'DJI', '262', '11.5', '43'),
    'Dominica': ('DM', 'DMA', '212', '15.4167', '-61.333'),
    'Dominican Republic': ('DO', 'DOM', '214', '19', '-70.666'),
    'Ecuador': ('EC', 'ECU', '218', '-2', '-77.5'),
    'Egypt': ('EG', 'EGY', '818', '27', '30'),
    'El Salvador': ('SV', 'SLV', '222', '13.833', '-88.916'),
    'Equatorial Guinea': ('GQ', 'GNQ', '226', '2', '10'),
    'Eritrea': ('ER', 'ERI', '232', '15', '39'),
    'Estonia': ('EE', 'EST', '233', '59', '26'),
    'Ethiopia': ('ET', 'ETH', '231', '8', '38'),
    'Falkland Islands Malvinas': ('FK', 'FLK', '238', '-51.75', '-59'),
    'Faroe Islands': ('FO', 'FRO', '234', '62', '-7'),
    'Fiji': ('FJ', 'FJI', '242', '-18', '175'),
    'Finland': ('FI', 'FIN', '246', '64', '26'),
    'France': ('FR', 'FRA', '250', '46', '2'),
    'French Guiana': ('GF', 'GUF', '254', '4', '-53'),
    'French Polynesia': ('PF', 'PYF', '258', '-15', '-140'),
    'French Southern Territories': ('TF', 'ATF', '260', '-43', '67'),
    'Gabon': ('GA', 'GAB', '266', '-1', '11.75'),
    'Gambia': ('GM', 'GMB', '270', '13.466', '-16.566'),
    'Georgia': ('GE', 'GEO', '268', '42', '43.5'),
    'Germany': ('DE', 'DEU', '276', '51', '9'),
    'Ghana': ('GH', 'GHA', '288', '8', '-2'),
    'Gibraltar': ('GI', 'GIB', '292', '36.183', '-5.366'),
    'Greece': ('GR', 'GRC', '300', '39', '22'),
    'Greenland': ('GL', 'GRL', '304', '72', '-40'),
    'Grenada': ('GD', 'GRD', '308', '12.116', '-61.666'),
    'Guadeloupe': ('GP', 'GLP', '312', '16.25', '-61.583'),
    'Guam': ('GU', 'GUM', '316', '13.4667', '144.783'),
    'Guatemala': ('GT', 'GTM', '320', '15.5', '-90.25'),
    'Guernsey': ('GG', 'GGY', '831', '49.5', '-2.56'),
    'Guinea': ('GN', 'GIN', '324', '11', '-10'),
    'Guinea-Bissau': ('GW', 'GNB', '624', '12', '-15'),
    'Guyana': ('GY', 'GUY', '328', '5', '-59'),
    'Haiti': ('HT', 'HTI', '332', '19', '-72.4167'),
    'Heard Island and McDonald Islands': ('HM', 'HMD', '334', '-53.1', '72.5167'),
    'Holy See': ('VA', 'VAT', '336', '41.9', '12.45'),
    'Vatican': ('VA', 'VAT', '336', '41.9', '12.45'),
    'Honduras': ('HN', 'HND', '340', '15', '-86.5'),
    'Hong Kong': ('HK', 'HKG', '344', '22.25', '114.1667'),
    'Hungary': ('HU', 'HUN', '348', '47', '20'),
    'Iceland': ('IS', 'ISL', '352', '65', '-18'),
    'India': ('IN', 'IND', '356', '20', '77'),
    'Indonesia': ('ID', 'IDN', '360', '-5', '120'),
    'Iran, Islamic Republic of': ('IR', 'IRN', '364', '32', '53'),
    'Iraq': ('IQ', 'IRQ', '368', '33', '44'),
    'Ireland': ('IE', 'IRL', '372', '53', '-8'),
    'Isle of Man': ('IM', 'IMN', '833', '54.23', '-4.55'),
    'Israel': ('IL', 'ISR', '376', '31.5', '34.75'),
    'Italy': ('IT', 'ITA', '380', '42.8333', '12.8333'),
    'Jamaica': ('JM', 'JAM', '388', '18.25', '-77.5'),
    'Japan': ('JP', 'JPN', '392', '36', '138'),
    'Jersey': ('JE', 'JEY', '832', '49.21', '-2.13'),
    'Jordan': ('JO', 'JOR', '400', '31', '36'),
    'Kazakhstan': ('KZ', 'KAZ', '398', '48', '68'),
    'Kenya': ('KE', 'KEN', '404', '1', '38'),
    'Kiribati': ('KI', 'KIR', '296', '1.4167', '173'),
    'Korea, Democratic Peoples Republic of': ('KP', 'PRK', '408', '40', '127'),
    'Korea, Republic of': ('KR', 'KOR', '410', '37', '127.5'),
    'South Korea': ('KR', 'KOR', '410', '37', '127.5'),
    'Kuwait': ('KW', 'KWT', '414', '29.3375', '47.6581'),
    'Kyrgyzstan': ('KG', 'KGZ', '417', '41', '75'),
    'Lao Peoples Democratic Republic': ('LA', 'LAO', '418', '18', '105'),
    'Latvia': ('LV', 'LVA', '428', '57', '25'),
    'Lebanon': ('LB', 'LBN', '422', '33.8333', '35.8333'),
    'Lesotho': ('LS', 'LSO', '426', '-29.5', '28.5'),
    'Liberia': ('LR', 'LBR', '430', '6.5', '-9.5'),
    'Libyan Arab Jamahiriya': ('LY', 'LBY', '434', '25', '17'),
    'Libya': ('LY', 'LBY', '434', '25', '17'),
    'Liechtenstein': ('LI', 'LIE', '438', '47.1667', '9.5333'),
    'Lithuania': ('LT', 'LTU', '440', '56', '24'),
    'Luxembourg': ('LU', 'LUX', '442', '49.75', '6.1667'),
    'Macao': ('MO', 'MAC', '446', '22.1667', '113.55'),
    'Macedonia, the former Yugoslav Republic of': ('MK', 'MKD', '807', '41.8333', '22'),
    'Madagascar': ('MG', 'MDG', '450', '-20', '47'),
    'Malawi': ('MW', 'MWI', '454', '-13.5', '34'),
    'Malaysia': ('MY', 'MYS', '458', '2.5', '112.5'),
    'Maldives': ('MV', 'MDV', '462', '3.25', '73'),
    'Mali': ('ML', 'MLI', '466', '17', '-4'),
    'Malta': ('MT', 'MLT', '470', '35.8333', '14.5833'),
    'Marshall Islands': ('MH', 'MHL', '584', '9', '168'),
    'Martinique': ('MQ', 'MTQ', '474', '14.6667', '-61'),
    'Mauritania': ('MR', 'MRT', '478', '20', '-12'),
    'Mauritius': ('MU', 'MUS', '480', '-20.2833', '57.55'),
    'Mayotte': ('YT', 'MYT', '175', '-12.8333', '45.1667'),
    'Mexico': ('MX', 'MEX', '484', '23', '-102'),
    'Micronesia, Federated States of': ('FM', 'FSM', '583', '6.9167', '158.25'),
    'Moldova, Republic of': ('MD', 'MDA', '498', '47', '29'),
    'Monaco': ('MC', 'MCO', '492', '43.7333', '7.4'),
    'Mongolia': ('MN', 'MNG', '496', '46', '105'),
    'Montenegro': ('ME', 'MNE', '499', '42', '19'),
    'Montserrat': ('MS', 'MSR', '500', '16.75', '-62.2'),
    'Morocco': ('MA', 'MAR', '504', '32', '-5'),
    'Mozambique': ('MZ', 'MOZ', '508', '-18.25', '35'),
    'Myanmar': ('MM', 'MMR', '104', '22', '98'),
    'Burma': ('MM', 'MMR', '104', '22', '98'),
    'Namibia': ('NA', 'NAM', '516', '-22', '17'),
    'Nauru': ('NR', 'NRU', '520', '-0.5333', '166.9167'),
    'Nepal': ('NP', 'NPL', '524', '28', '84'),
    'Netherlands': ('NL', 'NLD', '528', '52.5', '5.75'),
    'Netherlands Antilles': ('AN', 'ANT', '530', '12.25', '-68.75'),
    'New Caledonia': ('NC', 'NCL', '540', '-21.5', '165.5'),
    'New Zealand': ('NZ', 'NZL', '554', '-41', '174'),
    'Nicaragua': ('NI', 'NIC', '558', '13', '-85'),
    'Niger': ('NE', 'NER', '562', '16', '8'),
    'Nigeria': ('NG', 'NGA', '566', '10', '8'),
    'Niue': ('NU', 'NIU', '570', '-19.0333', '-169.8667'),
    'Norfolk Island': ('NF', 'NFK', '574', '-29.0333', '167.95'),
    'Northern Mariana Islands': ('MP', 'MNP', '580', '15.2', '145.75'),
    'Norway': ('NO', 'NOR', '578', '62', '10'),
    'Oman': ('OM', 'OMN', '512', '21', '57'),
    'Pakistan': ('PK', 'PAK', '586', '30', '70'),
    'Palau': ('PW', 'PLW', '585', '7.5', '134.5'),
    'Palestinian Territory, Occupied': ('PS', 'PSE', '275', '32', '35.25'),
    'Panama': ('PA', 'PAN', '591', '9', '-80'),
    'Papua New Guinea': ('PG', 'PNG', '598', '-6', '147'),
    'Paraguay': ('PY', 'PRY', '600', '-23', '-58'),
    'Peru': ('PE', 'PER', '604', '-10', '-76'),
    'Philippines': ('PH', 'PHL', '608', '13', '122'),
    'Pitcairn': ('PN', 'PCN', '612', '-24.7', '-127.4'),
    'Poland': ('PL', 'POL', '616', '52', '20'),
    'Portugal': ('PT', 'PRT', '620', '39.5', '-8'),
    'Puerto Rico': ('PR', 'PRI', '630', '18.25', '-66.5'),
    'Qatar': ('QA', 'QAT', '634', '25.5', '51.25'),
    'Réunion': ('RE', 'REU', '638', '-21.1', '55.6'),
    'Romania': ('RO', 'ROU', '642', '46', '25'),
    'Russia': ('RU', 'RUS', '643', '60', '100'),
    'Rwanda': ('RW', 'RWA', '646', '-2', '30'),
    'Saint Helena, Ascension and Tristan da Cunha': ('SH', 'SHN', '654', '-15.9333', '-5.7'),
    'Saint Kitts and Nevis': ('KN', 'KNA', '659', '17.3333', '-62.75'),
    'Saint Lucia': ('LC', 'LCA', '662', '13.8833', '-61.1333'),
    'Saint Pierre and Miquelon': ('PM', 'SPM', '666', '46.8333', '-56.3333'),
    'Saint Vincent and the Grenadines': ('VC', 'VCT', '670', '13.25', '-61.2'),
    'Samoa': ('WS', 'WSM', '882', '-13.5833', '-172.3333'),
    'San Marino': ('SM', 'SMR', '674', '43.7667', '12.4167'),
    'Sao Tome and Principe': ('ST', 'STP', '678', '1', '7'),
    'Saudi Arabia': ('SA', 'SAU', '682', '25', '45'),
    'Senegal': ('SN', 'SEN', '686', '14', '-14'),
    'Serbia': ('RS', 'SRB', '688', '44', '21'),
    'Seychelles': ('SC', 'SYC', '690', '-4.5833', '55.6667'),
    'Sierra Leone': ('SL', 'SLE', '694', '8.5', '-11.5'),
    'Singapore': ('SG', 'SGP', '702', '1.3667', '103.8'),
    'Slovakia': ('SK', 'SVK', '703', '48.6667', '19.5'),
    'Slovenia': ('SI', 'SVN', '705', '46', '15'),
    'Solomon Islands': ('SB', 'SLB', '90', '-8', '159'),
    'Somalia': ('SO', 'SOM', '706', '10', '49'),
    'South Africa': ('ZA', 'ZAF', '710', '-29', '24'),
    'South Georgia and the South Sandwich Islands': ('GS', 'SGS', '239', '-54.5', '-37'),
    'Spain': ('ES', 'ESP', '724', '40', '-4'),
    'Sri Lanka': ('LK', 'LKA', '144', '7', '81'),
    'Sudan': ('SD', 'SDN', '736', '15', '30'),
    'Suriname': ('SR', 'SUR', '740', '4', '-56'),
    'Svalbard and Jan Mayen': ('SJ', 'SJM', '744', '78', '20'),
    'Swaziland': ('SZ', 'SWZ', '748', '-26.5', '31.5'),
    'Sweden': ('SE', 'SWE', '752', '62', '15'),
    'Switzerland': ('CH', 'CHE', '756', '47', '8'),
    'Syrian Arab Republic': ('SY', 'SYR', '760', '35', '38'),
    'Taiwan, Province of China': ('TW', 'TWN', '158', '23.5', '121'),
    'Taiwan': ('TW', 'TWN', '158', '23.5', '121'),
    'Tajikistan': ('TJ', 'TJK', '762', '39', '71'),
    'Tanzania, United Republic of': ('TZ', 'TZA', '834', '-6', '35'),
    'Thailand': ('TH', 'THA', '764', '15', '100'),
    'Timor-Leste': ('TL', 'TLS', '626', '-8.55', '125.5167'),
    'Togo': ('TG', 'TGO', '768', '8', '1.1667'),
    'Tokelau': ('TK', 'TKL', '772', '-9', '-172'),
    'Tonga': ('TO', 'TON', '776', '-20', '-175'),
    'Trinidad and Tobago': ('TT', 'TTO', '780', '11', '-61'),
    'Tunisia': ('TN', 'TUN', '788', '34', '9'),
    'Turkey': ('TR', 'TUR', '792', '39', '35'),
    'Turkmenistan': ('TM', 'TKM', '795', '40', '60'),
    'Turks and Caicos Islands': ('TC', 'TCA', '796', '21.75', '-71.5833'),
    'Tuvalu': ('TV', 'TUV', '798', '-8', '178'),
    'Uganda': ('UG', 'UGA', '800', '1', '32'),
    'Ukraine': ('UA', 'UKR', '804', '49', '32'),
    'United Arab Emirates': ('AE', 'ARE', '784', '24', '54'),
    'United Kingdom': ('GB', 'GBR', '826', '54', '-2'),
    'United States': ('US', 'USA', '840', '38', '-97'),
    'U.S.A.': ('US', 'USA', '840', '38', '-97'),
    'USA': ('US', 'USA', '840', '38', '-97'),
    'United States Minor Outlying Islands': ('UM', 'UMI', '581', '19.2833', '166.6'),
    'Uruguay': ('UY', 'URY', '858', '-33', '-56'),
    'Uzbekistan': ('UZ', 'UZB', '860', '41', '64'),
    'Vanuatu': ('VU', 'VUT', '548', '-16', '167'),
    'Venezuela, Bolivarian Republic of': ('VE', 'VEN', '862', '8', '-66'),
    'Venezuela': ('VE', 'VEN', '862', '8', '-66'),
    'Vietnam': ('VN', 'VNM', '704', '16', '106'),
    'Virgin Islands, British': ('VG', 'VGB', '92', '18.5', '-64.5'),
    'Virgin Islands, U.S.': ('VI', 'VIR', '850', '18.3333', '-64.8333'),
    'Wallis and Futuna': ('WF', 'WLF', '876', '-13.3', '-176.2'),
    'Western Sahara': ('EH', 'ESH', '732', '24.5', '-13'),
    'Yemen': ('YE', 'YEM', '887', '15', '48'),
    'Zambia': ('ZM', 'ZMB', '894', '-15', '30'),
    'Zimbabwe': ('ZW', 'ZWE', '716', '-20', '30')
}

nl_trans = {
    "Afghanistan": "Afghanistan",
    "Albania": "Albanië",
    "Algeria": "Algerije",
    "Andorra": "Andorra",
    "Angola": "Angola",
    "Antigua and Barbuda": "Antigua en Barbuda",
    "Argentina": "Argentinië",
    "Armenia": "Armenië",
    "Australia": "Australië",
    "Austria": "Oostenrijk",
    "Azerbaijan": "Azerbeidzjan",
    "Bahamas": "Bahama’s",
    "Bahrain": "Bahrein",
    "Bangladesh": "Bangladesh",
    "Barbados": "Barbados",
    "Belarus": "Wit-Rusland",
    "Belgium": "België",
    "Belize": "Belize",
    "Benin": "Benin",
    "Bhutan": "Bhutan",
    "Bolivia": "Bolivia",
    "Bosnia and Herzegovina": "Bosnië en Herzegovina",
    "Botswana": "Botswana",
    "Brazil": "Brazilië",
    "Brunei": "Brunei",
    "Bulgaria": "Bulgarije",
    "Burkina Faso": "Burkina Faso",
    "Burundi": "Burundi",
    "Cambodia": "Cambodja",
    "Cameroon": "Kameroen",
    "Canada": "Canada",
    "Cape Verde": "Kaapverdië",
    "Central African Republic": "Centraal-Afrikaanse Republiek",
    "Chad": "Tsjaad",
    "Chile": "Chili",
    "China": "China",
    "Colombia": "Colombia",
    "Comoros": "Comoren",
    "Costa Rica": "Costa Rica",
    "Côte d’Ivoire": "Ivoorkust",
    "Croatia": "Kroatië",
    "Cuba": "Cuba",
    "Cyprus": "Cyprus",
    "Czech Republic": "Tsjechië",
    "Democratic Republic of the Congo": "Democratische Republiek Congo",
    "Denmark": "Denemarken",
    "Djibouti": "Djibouti",
    "Dominica": "Dominica",
    "Dominican Republic": "Dominicaanse Republiek",
    "East Timor": "Oost-Timor",
    "Ecuador": "Ecuador",
    "Egypt": "Egypte",
    "El Salvador": "El Salvador",
    "Equatorial Guinea": "Equatoriaal-Guinea",
    "Eritrea": "Eritrea",
    "Estonia": "Estland",
    "Ethiopia": "Ethiopië",
    "Fiji": "Fiji",
    "Finland": "Finland",
    "France": "Frankrijk",
    "Gabon": "Gabon",
    "Gambia": "Gambia",
    "Georgia": "Georgië",
    "Germany": "Duitsland",
    "Ghana": "Ghana",
    "Greece": "Griekenland",
    "Grenada": "Grenada",
    "Guatemala": "Guatemala",
    "Guinea": "Guinea",
    "Guinea-Bissau": "Guinee-Bissau",
    "Guyana": "Guyana",
    "Haiti": "Haïti",
    "Honduras": "Honduras",
    "Hungary": "Hongarije",
    "Iceland": "IJsland",
    "India": "Indië",
    "Indonesia": "Indonesië",
    "Iran": "Iran",
    "Iraq": "Irak",
    "Ireland": "Ierland",
    "Israel": "Israël",
    "Italy": "Italië",
    "Jamaica": "Jamaica",
    "Japan": "Japan",
    "Jordan": "Jordanië",
    "Kazakhstan": "Kazachstan",
    "Kenya": "Kenia",
    "Kiribati": "Kiribati",
    "Kuwait": "Koeweit",
    "Kyrgyzstan": "Kirgizië",
    "Laos": "Laos",
    "Latvia": "Letland",
    "Lebanon": "Libanon",
    "Lesotho": "Lesotho",
    "Liberia": "Liberia",
    "Libya": "Libië",
    "Liechtenstein": "Liechtenstein",
    "Lithuania": "Litouwen",
    "Luxembourg": "Luxemburg",
    "Madagascar": "Madagaskar",
    "Malawi": "Malawi",
    "Malaysia": "Maleisië",
    "Maldives": "Maldiven",
    "Mali": "Mali",
    "Malta": "Malta",
    "Marshall Islands": "Marshalleilanden",
    "Mauritania": "Mauritanië",
    "Mauritius": "Mauritius",
    "Mexico": "Mexico",
    "Micronesia": "Micronesië",
    "Moldova": "Moldavië",
    "Monaco": "Monaco",
    "Mongolia": "Mongolië",
    "Montenegro": "Montenegro",
    "Morocco": "Marokko",
    "Mozambique": "Mozambique",
    "Myanmar": "Myanmar",
    "Namibia": "Namibië",
    "Nauru": "Nauru",
    "Nepal": "Nepal",
    "Netherlands": "Nederland",
    "New Zealand": "Nieuw-Zeeland",
    "Nicaragua": "Nicaragua",
    "Niger": "Niger",
    "Nigeria": "Nigeria",
    "North Korea": "Noord-Korea",
    "Norway": "Noorwegen",
    "Oman": "Oman",
    "Pakistan": "Pakistan",
    "Palau": "Palau",
    "Panama": "Panama",
    "Papua New Guinea": "Papoea-Nieuw-Guinea",
    "Paraguay": "Paraguay",
    "Peru": "Peru",
    "Philippines": "Filipijnen",
    "Poland": "Polen",
    "Portugal": "Portugal",
    "Qatar": "Qatar",
    "Republic of the Congo": "Republiek Congo",
    "Republic of Macedonia": "Macedonië",
    "Romania": "Roemenië",
    "Russia": "Rusland",
    "Rwanda": "Rwanda",
    "Saint Kitts and Nevis": "Saint Kitts en Nevis",
    "Saint Lucia": "Saint Lucia",
    "Saint Vincent and the Grenadines": "Saint Vincent en de Grenadines",
    "Samoa": "Samoa",
    "San Marino": "San Marino",
    "Sao Tome and Principe": "Sao Tomé en Principe",
    "Saudi Arabia": "Saoedi-Arabië",
    "Senegal": "Senegal",
    "Serbia": "Servië",
    "Seychelles": "Seychellen",
    "Sierra Leone": "Sierra Leone",
    "Singapore": "Singapore",
    "Slovakia": "Slowakije",
    "Slovenia": "Slovenië",
    "Solomon Islands": "Salomonseilanden",
    "Somalia": "Somalië",
    "South Africa": "Zuid-Afrika",
    "South Korea": "Zuid-Korea",
    "South Sudan": "Zuid-Soedan",
    "Spain": "Spanje",
    "Sri Lanka": "Sri Lanka",
    "Sudan": "Soedan",
    "Suriname": "Suriname",
    "Swaziland": "Swaziland",
    "Sweden": "Zweden",
    "Switzerland": "Zwitserland",
    "Syria": "Syrië",
    "Tajikistan": "Tadzjikistan",
    "Tanzania": "Tanzania",
    "Thailand": "Thailand",
    "Togo": "Togo",
    "Tonga": "Tonga",
    "Trinidad and Tobago": "Trinidad en Tobago",
    "Tunisia": "Tunesië",
    "Turkey": "Turkije",
    "Turkmenistan": "Turkmenistan",
    "Tuvalu": "Tuvalu",
    "Uganda": "Oeganda",
    "Ukraine": "Oekraïne",
    "United Arab Emirates": "Verenigde Arabische Emiraten",
    "United Kingdom": "Verenigd Koninkrijk",
    "United States of America": "Verenigde Staten",
    "USA": "Verenigde Staten",
    "United States": "Verenigde Staten",
    "Uruguay": "Uruguay",
    "Uzbekistan": "Oezbekistan",
    "Vanuatu": "Vanuatu",
    "Venezuela": "Venezuela",
    "Vietnam": "Vietnam",
    "Yemen": "Jemen",
    "Zambia": "Zambia",
    "Zimbabwe": "Zimbabwe",
}

de_trans = {
    "Afghanistan": "Afghanistan",
    "Albania": "Albanien",
    "Algeria": "Algerien",
    "Andorra": "Andorra",
    "Angola": "Angola",
    "Antigua and Barbuda": "Antigua und Barbuda",
    "Argentina": "Argentinien",
    "Armenia": "Armenien",
    "Australia": "Australien",
    "Austria": "Österreich",
    "Azerbaijan": "Aserbaidschan",
    "Bahamas": "Bahamas",
    "Bahrain": "Bahrain",
    "Bangladesh": "Bangladesch",
    "Barbados": "Barbados",
    "Belarus": "Weißrussland",
    "Belgium": "Belgien",
    "Belize": "Belize",
    "Benin": "Benin",
    "Bhutan": "Bhutan",
    "Bolivia": "Bolivien",
    "Bosnia and Herzegovina": "Bosnien und Herzegowina",
    "Botswana": "Botswana",
    "Brazil": "Brasilien",
    "Brunei": "Brunei",
    "Bulgaria": "Bulgarien",
    "Burkina Faso": "Burkina Faso",
    "Burundi": "Burundi",
    "Cambodia": "Kambodscha",
    "Cameroon": "Kamerun",
    "Canada": "Kanada",
    "Cape Verde": "Kap Verde",
    "Central African Republic": "Zentralafrikanische Republik",
    "Chad": "Tschad",
    "Chile": "Chile",
    "China": "China",
    "Colombia": "Kolumbien",
    "Comoros": "Komoren",
    "Costa Rica": "Costa Rica",
    "Côte d’Ivoire": "Elfenbeinküste",
    "Croatia": "Kroatien",
    "Cuba": "Kuba",
    "Cyprus": "Zypern",
    "Czech Republic": "Tschechien",
    "Democratic Republic of the Congo": "Demokratische Republik Kongo",
    "Denmark": "Dänemark",
    "Djibouti": "Dschibuti",
    "Dominica": "Dominica",
    "Dominican Republic": "Dominikanische Republik",
    "East Timor": "Osttimor",
    "Ecuador": "Ecuador",
    "Egypt": "Ägypten",
    "El Salvador": "El Salvador",
    "Equatorial Guinea": "Äquatorialguinea",
    "Eritrea": "Eritrea",
    "Estonia": "Estland",
    "Ethiopia": "Äthiopien",
    "Fiji": "Fidschi",
    "Finland": "Finnland",
    "France": "Frankreich",
    "Gabon": "Gabun",
    "Gambia": "Gambia",
    "Georgia": "Georgia",
    "Germany": "Deutschland",
    "Ghana": "Ghana",
    "Greece": "Griechenland",
    "Grenada": "Grenada",
    "Guatemala": "Guatemala",
    "Guinea": "Guinea",
    "Guinea-Bissau": "Guinea-Bissau",
    "Guyana": "Guyana",
    "Haiti": "Haiti",
    "Honduras": "Honduras",
    "Hungary": "Ungarn",
    "Iceland": "Island",
    "India": "Indien",
    "Indonesia": "Indonesien",
    "Iran": "Iran",
    "Iraq": "Irak",
    "Ireland": "Irland",
    "Israel": "Israel",
    "Italy": "Italien",
    "Jamaica": "Jamaika",
    "Japan": "Japan",
    "Jordan": "Jordan",
    "Kazakhstan": "Kasachstan",
    "Kenya": "Kenia",
    "Kiribati": "Kiribati",
    "Kuwait": "Kuwait",
    "Kyrgyzstan": "Kirgisistan",
    "Laos": "Laos",
    "Latvia": "Lettland",
    "Lebanon": "Libanon",
    "Lesotho": "Lesotho",
    "Liberia": "Liberia",
    "Libya": "Libyen",
    "Liechtenstein": "Liechtenstein",
    "Lithuania": "Litauen",
    "Luxembourg": "Luxemburg",
    "Madagascar": "Madagaskar",
    "Malawi": "Malawi",
    "Malaysia": "Malaysia",
    "Maldives": "Malediven",
    "Mali": "Mali",
    "Malta": "Malta",
    "Marshall Islands": "Marshallinseln",
    "Mauritania": "Mauretanien",
    "Mauritius": "Mauritius",
    "Mexico": "Mexiko",
    "Micronesia": "Mikronesien",
    "Moldova": "Moldawien",
    "Monaco": "Monaco",
    "Mongolia": "Mongolei",
    "Montenegro": "Montenegro",
    "Morocco": "Marokko",
    "Mozambique": "Mosambik",
    "Myanmar": "Myanmar",
    "Namibia": "Namibia",
    "Nauru": "Nauru",
    "Nepal": "Nepal",
    "Netherlands": "Niederlande",
    "New Zealand": "Neuseeland",
    "Nicaragua": "Nicaragua",
    "Niger": "Niger",
    "Nigeria": "Nigeria",
    "North Korea": "Nordkorea",
    "Norway": "Norwegen",
    "Oman": "Oman",
    "Pakistan": "Pakistan",
    "Palau": "Palau",
    "Panama": "Panama",
    "Papua New Guinea": "Papua-Neuguinea",
    "Paraguay": "Paraguay",
    "Peru": "Peru",
    "Philippines": "Philippinen",
    "Poland": "Polen",
    "Portugal": "Portugal",
    "Qatar": "Katar",
    "Republic of the Congo": "Republik Kongo",
    "Republic of Macedonia": "Mazedonien",
    "Romania": "Rumänien",
    "Russia": "Russland",
    "Rwanda": "Ruanda",
    "Saint Kitts and Nevis": "St. Kitts und Nevis",
    "Saint Lucia": "St. Lucia",
    "Saint Vincent and the Grenadines": "St. Vincent und die Grenadinen",
    "Samoa": "Samoa",
    "San Marino": "San Marino",
    "Sao Tome and Principe": "São Tomé und Príncipe",
    "Saudi Arabia": "Saudi-Arabien",
    "Senegal": "Senegal",
    "Serbia": "Serbien",
    "Seychelles": "Seychellen",
    "Sierra Leone": "Sierra Leone",
    "Singapore": "Singapur",
    "Slovakia": "Slowakei",
    "Slovenia": "Slowenien",
    "Solomon Islands": "Salomonen",
    "Somalia": "Somalia",
    "South Africa": "Südafrika",
    "South Korea": "Südkorea",
    "South Sudan": "Südsudan",
    "Spain": "Spanien",
    "Sri Lanka": "Sri Lanka",
    "Sudan": "Sudan",
    "Suriname": "Suriname",
    "Swaziland": "Swasiland",
    "Sweden": "Schweden",
    "Switzerland": "Schweiz",
    "Syria": "Syrien",
    "Tajikistan": "Tadschikistan",
    "Tanzania": "Tansania",
    "Thailand": "Thailand",
    "Togo": "Togo",
    "Tonga": "Tonga",
    "Trinidad and Tobago": "Trinidad und Tobago Tobago",
    "Tunisia": "Tunesien",
    "Turkey": "Türkei",
    "Turkmenistan": "Turkmenistan",
    "Tuvalu": "Tuvalu",
    "Uganda": "Uganda",
    "Ukraine": "Ukraine",
    "United Arab Emirates": "Vereinigte Arabische Emirate Emirate",
    "United Kingdom": "Königreich Großbritannien",
    "United States of America": "Vereinigte Staaten",
    "USA": "Vereinigte Staaten",
    "United States": "Vereinigte Staaten",
    "Uruguay": "Uruguay",
    "Uzbekistan": "Usbekistan",
    "Vanuatu": "Vanuatu",
    "Venezuela": "Venezuela",
    "Vietnam": "Vietnam",
    "Yemen": "Jemen",
    "Zambia": "Sambia",
    "Zimbabwe": "Simbabwe",

}

trans_table = {
    'nl': nl_trans,
    #'de': de_trans,
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
import copy
import logging
import math
import re
import string as st

from geofinder import Loc


class CSVEntry:
    PLACE_ID = 0
    TITLE = 1
    NAME = 2
    LAT = 3
    LON = 4
    FEAT = 5
    ADMIN1_ID = 6
    ADMIN2_ID = 7
    ISO = 8
    ENCLOSED_BY = 9
    TYPE = 10


class GrampsCsv:
    def __init__(self, in_path: str, geodata):
        self.logger = logging.getLogger(__name__)
        self.csv_path = in_path + '.' + 'csv'

        self.admin_table = [{}, {}, {}, {}, {}]
        self.csvfile = None
        self.geodata = geodata

    @staticmethod
    def get_dict_id(place):
        if place.place_type == Loc.PlaceType.COUNTRY:
            dict_idx = 0
        elif place.place_type == Loc.PlaceType.ADMIN1:
            dict_idx = 1
        elif place.place_type == Loc.PlaceType.ADMIN2:
            dict_idx = 2
        elif place.place_type == Loc.PlaceType.CITY:
            dict_idx = 3
        elif place.place_type == Loc.PlaceType.PREFIX:
            dict_idx = 4
        else:
            msg = f'Dictionary ID - Unknown place type for {place.name}. Type={place.place_type}'
            raise Exception(msg)
        return dict_idx

    def write_asis(self, entry:str):
        pass

    def get_csv_key(self, place):
        # Fill in admin2
        if place.admin2_id == '':
            self.geodata.geo_files.geodb.get_admin2_id(place)
        if place.admin2_id == '' and len(place.admin2_name.strip(' ')) > 0:
            place.admin2_id = ' '
        if place.place_type == Loc.PlaceType.COUNTRY:
            key = f'{place.country_iso}'
        elif place.place_type == Loc.PlaceType.ADMIN1:
            key = f'{place.admin1_id}_{place.country_iso}'
        elif place.place_type == Loc.PlaceType.ADMIN2:
            key = f'{place.admin2_id}_{place.admin1_id}_{place.country_iso}'
        elif place.place_type == Loc.PlaceType.CITY:
            key = f'{place.city1.strip(" ")}_{place.admin2_id}_{place.admin1_id}_{place.country_iso}'
        elif place.place_type == Loc.PlaceType.PREFIX:
            key = f'{place.prefix.strip(" ")}_{place.city1.strip(" ")}_{place.admin2_id}_{place.admin1_id}_{place.country_iso}'
        else:
            msg = f'Get key - Unknown place type for {place.name}. Type={place.place_type}'
            raise Exception(msg)

        key = key.strip('_')
        key = key.strip('_')
        key = key.strip(' ')
        # self.logger.debug(f'key={key.upper().strip("_")} type={place.place_type}')
        return key.upper()

    @staticmethod
    def set_CSV_place_type(place: Loc.Loc):
        place.set_place_type()
        if len(place.prefix) > 0:
            place.place_type = Loc.PlaceType.PREFIX

    @staticmethod
    def get_csv_name(place):
        if place.place_type == Loc.PlaceType.COUNTRY:
            place.name = place.country_name
            nm = place.country_name
        elif place.place_type == Loc.PlaceType.ADMIN1:
            nm = place.admin1_name
        elif place.place_type == Loc.PlaceType.ADMIN2:
            nm = place.admin2_name
        elif place.place_type == Loc.PlaceType.CITY:
            nm = place.city1
        elif place.place_type == Loc.PlaceType.PREFIX:
            nm = place.prefix
        else:
            msg = f'Get name - Unknown place type for {place.name}. Type={place.place_type}'
            raise Exception(msg)
        return nm

    def create_csv_node(self, place: Loc.Loc):
        """
        Create CSV row in Dictionary:  Place (ID), Title, Name, Type, latitude, longitude,enclosed_by
        :param place:
        :return: None
        """
        if place.name == '':
            return

        row = [''] * 11
        self.set_CSV_place_type(place)

        if place.id == '':
            self.set_CSV_place_type(place)
            place.id = self.get_csv_key(place)

        row[CSVEntry.PLACE_ID] = place.id
        row[CSVEntry.ENCLOSED_BY] = place.enclosed_by
        place.id = row[CSVEntry.PLACE_ID]

        row[CSVEntry.TITLE] = place.prefix + place.prefix_commas + place.name
        #row[CSVEntry.TITLE] = re.sub("S ", "s ", row[CSVEntry.TITLE])  # Fix the apostrophe S problem in Titles
        row[CSVEntry.FEAT] = place.feature
        row[CSVEntry.LAT] = f'{float(place.lat):.4f}'

        row[CSVEntry.LAT] = f'{float(place.lat):.4f}'
        row[CSVEntry.LON] = f'{float(place.lon):.4f}'
        row[CSVEntry.ADMIN2_ID] = place.admin2_id
        row[CSVEntry.ADMIN1_ID] = place.admin1_id
        row[CSVEntry.ISO] = place.country_iso

        #place.calculate_type()
        self.geodata.set_place_type_text(place)
        row[CSVEntry.NAME] = self.get_csv_name(place)
        #row[CSVEntry.NAME] = re.sub("S ", "s ", row[CSVEntry.NAME])  # Fix the apostrophe S problem in Titles
        row[CSVEntry.TYPE] = place.result_type_text
        key = self.get_csv_key(place)
        dict_idx = self.get_dict_id(place)

        if dict_idx == 0:
            place.enclosed_by = ''
            row[CSVEntry.ENCLOSED_BY] = ''

        if place.enclosed_by != '':
            if key.count('_') <= row[CSVEntry.ENCLOSED_BY].count('_') and key.count('_') > 0:
                msg = f'Incorrect Enclosure for [{place.name}]. Key= [{key}] Enclosure= [{row[CSVEntry.ENCLOSED_BY]}]'
                self.logger.warning(msg)
            elif key.count('_') < row[CSVEntry.ENCLOSED_BY].count('_') and key.count('_') == 0:
                msg = f'Incorrect Enclosure for [{place.name}]. Key= [{key}] Enclosure= [{row[CSVEntry.ENCLOSED_BY]}]'
                self.logger.warning(msg)

        if re.match(r'P\d\d\d\d', place.id):
            # our item has an ID with P9999,  add this row
            self.admin_table[dict_idx][key.upper()] = row
        else:
            res = self.admin_table[dict_idx].get(key.upper())
            if res is None:
                # Nothing there, add this row
                self.admin_table[dict_idx][key.upper()] = row
            else:
                # A node is already there and we don't have a P, so do nothing
                place.id = res[CSVEntry.PLACE_ID]

        self.logger.debug(f'\nCREATE CSV NODE {key.upper()} idx={dict_idx}: {row}\n{place.name}')

    def move_up_level(self, enclosure_place, idx) -> bool:
        enclosure_place.lat = 99.9
        enclosure_place.lon = 99.9

        # Switch place type to next level higher
        if enclosure_place.place_type == Loc.PlaceType.COUNTRY:
            # Already at top
            enclosure_place.feature = 'ADM0'
            return False
        elif enclosure_place.place_type == Loc.PlaceType.ADMIN1:
            enclosure_place.feature = 'ADM1'
            enclosure_place.place_type = Loc.PlaceType.COUNTRY
        elif enclosure_place.place_type == Loc.PlaceType.ADMIN2:
            enclosure_place.feature = 'ADM2'
            enclosure_place.place_type = Loc.PlaceType.ADMIN1
        elif enclosure_place.place_type == Loc.PlaceType.CITY:
            enclosure_place.place_type = Loc.PlaceType.ADMIN2
        elif enclosure_place.place_type == Loc.PlaceType.PREFIX:
            enclosure_place.place_type = Loc.PlaceType.CITY
        else:
            msg = f'Move Up - Unknown place type for {enclosure_place.name}. Type={enclosure_place.place_type}'
            raise Exception(msg)

        enclosure_place.remove_old_fields()
        enclosure_place.name = enclosure_place.format_full_nm(None)
        self.geodata.set_place_type_text(enclosure_place)
        # enclosure_place.city1 = tkns[1]
        save_type = enclosure_place.place_type
        self.geodata.find_first_match(enclosure_place.name, enclosure_place)
        enclosure_place.place_type = save_type
        enclosure_place.remove_old_fields()
        enclosure_place.name = enclosure_place.format_full_nm(None)
        self.geodata.set_place_type_text(enclosure_place)
        enclosure_place.id = self.get_csv_key(enclosure_place)

        # place.name = place.format_full_nm(None)
        self.logger.debug(f'\nMOVED UP TO {enclosure_place.name}')

        self.create_csv_node(enclosure_place)

        new_idx = self.get_dict_id(enclosure_place)
        if new_idx < idx:
            return True
        else:
            msg = f'Move Up - Index error {enclosure_place.name}. Type={enclosure_place.place_type} idx={idx} new_idx={new_idx}'
            raise Exception(msg)

    def create_enclosed_by(self, place: Loc.Loc):
        """
        Create EnclosedBy elements in Dictionary for CSV file
        :return: None
        """
        self.logger.debug(f'\nCREATE ENCLOSURE FOR {place.name}')
        enclosure_place = copy.copy(place)
        enclosure_place.id = ''

        # Move up to enclosure level
        success = self.move_up_level(enclosure_place, idx=self.get_dict_id(enclosure_place))
        if success:
            place.enclosed_by = enclosure_place.id
            self.update_enclosure_id(place)
        return

    def update_enclosure_id(self, place):
        key = self.get_csv_key(place)
        dict_idx = self.get_dict_id(place)
        row = self.admin_table[dict_idx].get(key.upper())
        if row:
            if not re.match(r'P\d\d\d\d', row[CSVEntry.ENCLOSED_BY]):
                row[CSVEntry.ENCLOSED_BY] = place.enclosed_by
                self.admin_table[dict_idx][key.upper()] = row
                self.logger.debug(f'UPDATE ENC for {dict_idx}:{key} New Enclosure=[{place.enclosed_by}]')
            else:
                pass
        else:
            self.logger.warning(f'@@@@@@@@ Cant find row {key}')

    def complete_csv(self):
        # Add location enclosures.  Create if not there already.  Then add as reference.
        self.logger.debug('\n\n******** DONE - CREATE CSV ENCLOSURES *********')
        place = Loc.Loc()

        # There are separate dictionaries for each hierarchy (prefix, city, county, country).
        # We need to go through prefix table, then city, etc (e.g. reversed order)
        # Create Enclosure records
        for idx, tbl in reversed(list(enumerate(self.admin_table))):
            self.logger.debug(f'===TABLE {idx}===')
            for key in tbl:
                self.retrieve_csv_place(self.admin_table, self.geodata, place, key, idx)
                self.logger.debug(f'** CSV {key} {place.name}')

                # Create enclosure for each node at this level
                self.create_enclosed_by(place)

        if self.csv_path is not None:
            self.csvfile = open(self.csv_path, "w", encoding='utf-8')
            self.logger.debug(f'CSV file {self.csv_path}')
            self.csvfile.write('Place,Title,Name,Type,latitude,longitude,enclosed_by\n')

        # List CSV
        self.logger.debug('*** OUTPUT TABLE ***')
        for idx, tbl in enumerate(self.admin_table):
            for key in tbl:
                # TODO
                row = tbl[key]
                self.logger.debug(f'IDX={idx} {key} : {row}')
                self.output_row(row)

        if self.csv_path is not None:
            self.csvfile.close()

    def output_row(self, row):
        if len(row[CSVEntry.ENCLOSED_BY]) > 0:
            enc = f'[{row[CSVEntry.ENCLOSED_BY]}]'
        else:
            enc = ''
        if self.csv_path is not None:
            # 0Place (ID), 1Title, 2Name, 3Type, 4latitude, 5longitude, 6enclosed_by
            title = st.capwords(row[CSVEntry.TITLE])
            title = re.sub("S ", "s ", title)  # Fix the apostrophe S problem in Titles

            name = st.capwords(row[CSVEntry.NAME])
            name = re.sub("S ", "s ", name)  # Fix the apostrophe S problem in Titles

            if math.isnan(float(row[CSVEntry.LAT])) or  math.isnan(float(row[CSVEntry.LAT])) :
                self.csvfile.write(f'[{row[CSVEntry.PLACE_ID]}],"{title}","{name}",{row[CSVEntry.TYPE]},'
                               f' , ,{enc},\n')
            else:
                self.csvfile.write(f'[{row[CSVEntry.PLACE_ID]}],"{title}","{name}",{row[CSVEntry.TYPE]},'
                               f'{row[CSVEntry.LAT]},{row[CSVEntry.LON]},{enc},\n')

    @staticmethod
    def retrieve_csv_place(admin_table, geodata, place: Loc.Loc, key, idx):
        # 0Place (ID), 1Title, 2Name, 3Type, 4latitude, 5longitude, 6enclosed_by
        row = admin_table[idx].get(key)
        key_tokens = key.split("_")
        place.place_type = len(key_tokens) - 1
        # self.logger.debug(f'{row}')
        place.feature = row[CSVEntry.FEAT]

        place.name = row[CSVEntry.TITLE]
        place.country_iso = row[CSVEntry.ISO]
        place.country_name = geodata.geo_files.geodb.get_country_name(place.country_iso)
        place.enclosed_by = row[CSVEntry.ENCLOSED_BY]

        place.lat: float = float(row[CSVEntry.LAT])
        place.lon: float = float(row[CSVEntry.LON])

        place.admin2_id = row[CSVEntry.ADMIN2_ID]
        place.admin1_id = row[CSVEntry.ADMIN1_ID]
        place.admin1_name = str(geodata.geo_files.geodb.get_admin1_name(place))
        place.admin2_name = str(geodata.geo_files.geodb.get_admin2_name(place))
        if place.admin2_name is None:
            place.admin2_name = ''
        if place.admin1_name is None:
            place.admin1_name = ''

        tokens = place.name.split(',')
        if len(tokens) > 3:
            place.city1 = tokens[-4]
        if len(tokens) > 4:
            place.prefix = tokens[-5]

        place.id = row[CSVEntry.PLACE_ID]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2019.       Mike Herbert
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

import logging
import os
import unittest
from pathlib import Path

from geofinder import Geodata, Loc, GrampsXml, GrampsCsv


#             ('12 Privet Drive,Dover, ,England,United Kingdom', "PPL", 'P0006', 'gb'),
class RowEntry:
    NAME = 0
    FEAT = 1
    PLACE_ID = 2
    ISO = 3


class TestCSV(unittest.TestCase):
    geodata = None
    ancestry = None
    logger = None
    csv = None

    @classmethod
    def setUpClass(cls):
        TestCSV.logger = logging.getLogger(__name__)
        fmt = "%(levelname)s %(name)s.%(funcName)s %(lineno)d: %(message)s"
        logging.basicConfig(level=logging.DEBUG, format=fmt)

        # Load test data
        directory = os.path.join(str(Path.home()), "geoname_test")
        csv_path = os.path.join(directory, "test")
        TestCSV.geodata = Geodata.Geodata(directory_name=directory, progress_bar=None)
        error: bool = TestCSV.geodata.read()
        if error:
            TestCSV.logger.error("Missing geodata support Files.")
            raise ValueError('Cannot open database')

        # Read in Geoname Gazeteer file - city names, lat/long, etc.
        error = TestCSV.geodata.read_geonames()

        if error:
            TestCSV.logger.info("Missing geoname Files.")
            TestCSV.logger.info('Requires ca.txt, gb.txt, de.txt from geonames.org in folder username/geoname_test')
            raise ValueError('Cannot open database')

        TestCSV.csv = GrampsCsv.GrampsCsv(in_path=csv_path, geodata=TestCSV.geodata)

        TestCSV.ancestry = GrampsXml.GrampsXml(in_path=csv_path, out_suffix='', cache_d=None,
                                               progress=None, geodata=TestCSV.geodata)  # Routines

        # Set up CSV Data
        csv_data = [
            #('Portugal', 'ADM0', 'P0001', 'po'),
             #('Scotland,United Kingdom', 'ADM1','P0002','gb'),
             #('Kent,England,United Kingdom', 'ADM2','P0003','gb'),
            #('Canterbury,Kent,England,United Kingdom', "PPL", 'P0004', 'gb'),
            #('St Eustace,Canterbury,Kent,England,United Kingdom', "PPL", 'P0005', 'gb'),
            #('Dover, ,England,United Kingdom', "PPL", 'P0006', 'gb'),
            ('12 Privet Drive, Dover, ,England,United Kingdom', "PPL", 'P0006', 'gb'),
            #('Edinburgh, ,Scotland,United Kingdom', "PPL", 'P0007', 'gb'),
            #("St James's Palace, ,England,United Kingdom", "PPL", 'P0008', 'gb'),
        ]

        place = Loc.Loc()

        for row in csv_data:
            place.clear()
            place.name = row[RowEntry.NAME]
            place.feature = row[RowEntry.FEAT]
            place.parse_place(place_name=place.name, geo_files=TestCSV.geodata.geo_files)
            # Lookup record
            TestCSV.geodata.find_first_match(place.name, place)
            place.id = row[RowEntry.PLACE_ID]
            TestCSV.csv.set_CSV_place_type(place)
            #TestCSV.geodata.set_place_type_text(place)
            #place.name = TestCSV.ancestry.get_csv_name(place).title()

            TestCSV.csv.create_csv_node(place)

        TestCSV.csv.complete_csv()

    def setUp(self) -> None:
        self.place: Loc.Loc = Loc.Loc()

    def run_key_test(self, title: str, entry: str):
        print("*****TEST: {}".format(title))
        place = Loc.Loc()
        place.parse_place(entry, TestCSV.geodata.geo_files )
        self.geodata.find_first_match(place.name, place)
        place.name = place.format_full_nm(None)
        TestCSV.csv.set_CSV_place_type(place)
        place.id = TestCSV.csv.get_csv_key(place)

        TestCSV.logger.debug(f'type={place.place_type}')

        TestCSV.geodata.set_place_type_text(place)
        #place_name = TestCSV.ancestry_file_handler.get_csv_name(place).title()
        return place.id

    # ===== TEST RESULT CODES

    def test_key01(self):
        title = "City.  Good.  upper lowercase"
        key = self.run_key_test(title, "Dover,Kent,England,United Kingdom")
        self.assertEqual("DOVER_G5_ENG_GB", key, title)
    
    def test_key02(self):
        title = "City.  Good.  upper lowercase"
        key = self.run_key_test(title, "Kent,England,United Kingdom")
        self.assertEqual("G5_ENG_GB", key, title)

    def test_key03(self):
        title = "City.  Good.  upper lowercase"
        key = self.run_key_test(title, "England,United Kingdom")
        self.assertEqual("ENG_GB", key, title)

    def test_key04(self):
        title = "City.  Good.  upper lowercase"
        key = self.run_key_test(title, "United Kingdom")
        self.assertEqual("GB", key, title)

    def test_key05(self):
        title = "City.  Good.  upper lowercase"
        key = self.run_key_test(title, "Scotland,United Kingdom")
        self.assertEqual("SCT_GB", key, title)

    def test_key06(self):
        title = "City.  Good.  upper lowercase"
        key = self.run_key_test(title, "Dover,xyz,England,United Kingdom")
        self.assertEqual("DOVER_G5_ENG_GB", key, title)

    def test_key07(self):
        title = "City.  Good.  upper lowercase"
        key = self.run_key_test(title, "Dover, ,England,United Kingdom")
        self.assertEqual("DOVER_G5_ENG_GB", key, title)

    def test_key08(self):
        title = "City.  Good.  upper lowercase"
        key = self.run_key_test(title, "Marks Square,Dover,,England,United Kingdom")
        self.assertEqual("MARKS SQUARE_DOVER_G5_ENG_GB", key, title)

    def test_key09(self):
        title = "City.  Good.  upper lowercase"
        key = self.run_key_test(title, "St. Andrews,Charlotte County,new brunswick,Canada")
        self.assertEqual("ST ANDREWS_1302_04_CA", key, title)


if __name__ == '__main__':
    unittest.main()
